# stdlib
import warnings

# its not clear what is causing this warning spam but it must be a third party numerical
# library changing numpy's smallest subnormal or something?
# https://github.com/numpy/numpy/issues/20895
warnings.filterwarnings(
    "ignore",
    ".*The value of the smallest subnormal for.*",
    category=UserWarning,
)

# DeprecationWarning: 'urllib3.contrib.pyopenssl' module is deprecated and will be
# removed in a future release of urllib3 2.x.
# Read more in this issue: https://github.com/urllib3/urllib3/issues/2680
warnings.filterwarnings(
    "ignore",
    ".*urllib3.contrib.pyopenssl.*",
    category=DeprecationWarning,
)

# UserWarning: libuv only supports millisecond timer resolution; all times less will be
# # set to 1 ms
warnings.filterwarnings(
    "ignore",
    ".*libuv only supports millisecond timer resolution*",
    category=UserWarning,
)

# site-packages/IPython/core/formatters.py:367: FormatterWarning:
#  text/html formatter returned invalid type <class 'type'> (expected <class 'str'>)
#  for object: <EnclaveClient - cranky_norvig <a2fa3fe86007429189dcd88081185bd5>: HTTPConnection: http://localhost:8082>
# during printing of client object
warnings.filterwarnings(
    "ignore",
    ".*text/html formatter returned invalid type <class 'type'>*",
    category=UserWarning,
)



# stdlib
from typing import List
from typing import Union

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import ActionObjectREAD
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from .notification_stash import NotificationStash
from .notifications import CreateNotification
from .notifications import LinkedObject
from .notifications import Notification
from .notifications import NotificationStatus
from .notifications import ReplyNotification


@instrument
@serializable()
class NotificationService(AbstractService):
    store: DocumentStore
    stash: NotificationStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = NotificationStash(store=store)

    @service_method(path="notifications.send", name="send")
    def send(
        self, context: AuthedServiceContext, notification: CreateNotification
    ) -> Union[Notification, SyftError]:
        """Send a new notification"""

        new_notification = notification.to(Notification, context=context)

        # Add read permissions to person receiving this message
        permissions = [
            ActionObjectREAD(
                uid=new_notification.id, credentials=new_notification.to_user_verify_key
            )
        ]

        result = self.stash.set(
            context.credentials, new_notification, add_permissions=permissions
        )

        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="notifications.reply", name="reply", roles=GUEST_ROLE_LEVEL)
    def reply(
        self,
        context: AuthedServiceContext,
        reply: ReplyNotification,
    ) -> Union[ReplyNotification, SyftError]:
        msg = self.stash.get_by_uid(
            credentials=context.credentials, uid=reply.target_msg
        )
        if msg.is_ok():
            msg = msg.ok()
            reply.from_user_verify_key = context.credentials
            msg.replies.append(reply)
            result = self.stash.update(credentials=context.credentials, obj=msg)
            if result.is_ok():
                return result.ok()
            else:
                SyftError(
                    message="Couldn't add a new notification reply in the target notification."
                )
        else:
            SyftError(
                message="The target notification id {reply.target_msg} was not found!"
            )

    @service_method(
        path="notifications.get_all",
        name="get_all",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all(
        self,
        context: AuthedServiceContext,
    ) -> Union[List[Notification], SyftError]:
        result = self.stash.get_all_inbox_for_verify_key(
            context.credentials,
            verify_key=context.credentials,
        )
        if result.err():
            return SyftError(message=str(result.err()))
        notifications = result.ok()
        return notifications

    @service_method(
        path="notifications.get_all_sent",
        name="outbox",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_sent(
        self, context: AuthedServiceContext
    ) -> Union[List[Notification], SyftError]:
        result = self.stash.get_all_sent_for_verify_key(
            context.credentials, context.credentials
        )
        if result.err():
            return SyftError(message=str(result.err()))
        notifications = result.ok()
        return notifications

    # get_all_read and unread cover the same functionality currently as
    # get_all_for_status. However, there may be more statuses added in the future,
    # so we are keeping the more generic get_all_for_status method.
    def get_all_for_status(
        self,
        context: AuthedServiceContext,
        status: NotificationStatus,
    ) -> Union[List[Notification], SyftError]:
        result = self.stash.get_all_by_verify_key_for_status(
            context.credentials, verify_key=context.credentials, status=status
        )
        if result.err():
            return SyftError(message=str(result.err()))
        notifications = result.ok()
        return notifications

    @service_method(
        path="notifications.get_all_read",
        name="get_all_read",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_read(
        self,
        context: AuthedServiceContext,
    ) -> Union[List[Notification], SyftError]:
        return self.get_all_for_status(
            context=context,
            status=NotificationStatus.READ,
        )

    @service_method(
        path="notifications.get_all_unread",
        name="get_all_unread",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_unread(
        self,
        context: AuthedServiceContext,
    ) -> Union[List[Notification], SyftError]:
        return self.get_all_for_status(
            context=context,
            status=NotificationStatus.UNREAD,
        )

    @service_method(path="notifications.mark_as_read", name="mark_as_read")
    def mark_as_read(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[Notification, SyftError]:
        result = self.stash.update_notification_status(
            context.credentials, uid=uid, status=NotificationStatus.READ
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="notifications.mark_as_unread", name="mark_as_unread")
    def mark_as_unread(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[Notification, SyftError]:
        result = self.stash.update_notification_status(
            context.credentials, uid=uid, status=NotificationStatus.UNREAD
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(
        path="notifications.resolve_object",
        name="resolve_object",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def resolve_object(
        self, context: AuthedServiceContext, linked_obj: LinkedObject
    ) -> Union[Notification, SyftError]:
        service = context.node.get_service(linked_obj.service_type)
        result = service.resolve_link(context=context, linked_obj=linked_obj)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="notifications.clear", name="clear")
    def clear(self, context: AuthedServiceContext) -> Union[SyftError, SyftSuccess]:
        result = self.stash.delete_all_for_verify_key(
            credentials=context.credentials, verify_key=context.credentials
        )
        if result.is_ok():
            return SyftSuccess(message="All notifications cleared !!")
        return SyftError(message=str(result.err()))

    def filter_by_obj(
        self, context: AuthedServiceContext, obj_uid: UID
    ) -> Union[Notification, SyftError]:
        notifications = self.stash.get_all(context.credentials)
        if notifications.is_ok():
            for notification in notifications.ok():
                if (
                    notification.linked_obj
                    and notification.linked_obj.object_uid == obj_uid
                ):
                    return notification
        else:
            return SyftError(message="Could not get notifications!!")


TYPE_TO_SERVICE[Notification] = NotificationService
SERVICE_TO_TYPES[NotificationService].update({Notification})



# stdlib
from typing import Union

# third party
from result import Ok

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .log import SyftLogV2
from .log_stash import LogStash


@instrument
@serializable()
class LogService(AbstractService):
    store: DocumentStore
    stash: LogStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = LogStash(store=store)

    @service_method(path="log.add", name="add", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def add(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        new_log = SyftLogV2(id=uid)
        result = self.stash.set(context.credentials, new_log)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result

    @service_method(path="log.append", name="append", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def append(
        self,
        context: AuthedServiceContext,
        uid: UID,
        new_str: str = "",
        new_err: str = "",
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))
        new_log = result.ok()
        if new_str:
            new_log.append(new_str)

        if new_err:
            new_log.append_error(new_err)

        result = self.stash.update(context.credentials, new_log)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="Log Append successful!")

    @service_method(path="log.get", name="get", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def get(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))

        return Ok(result.ok().stdout)

    @service_method(path="log.restart", name="restart", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def restart(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))

        log = result.ok()
        log.restart()
        result = self.stash.update(context.credentials, log)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="Log Restart successful!")

    @service_method(path="log.get_error", name="get_error", roles=ADMIN_ROLE_LEVEL)
    def get_error(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))

        return Ok(result.ok().stderr)

    @service_method(path="log.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="log.delete", name="delete", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.delete_by_uid(context.credentials, uid)
        if result.is_ok():
            return result.ok()
        else:
            return SyftError(message=result.err())



# stdlib
from pathlib import Path
from typing import List
from typing import Optional
from typing import Union

# third party
import requests

# relative
from ...serde.serializable import serializable
from ...service.action.action_object import ActionObject
from ...store.blob_storage import BlobRetrieval
from ...store.blob_storage.on_disk import OnDiskBlobDeposit
from ...store.blob_storage.seaweedfs import SeaweedFSBlobDeposit
from ...store.document_store import DocumentStore
from ...store.document_store import UIDPartitionKey
from ...types.blob_storage import AzureSecureFilePathLocation
from ...types.blob_storage import BlobFileType
from ...types.blob_storage import BlobStorageEntry
from ...types.blob_storage import BlobStorageMetadata
from ...types.blob_storage import CreateBlobStorageEntry
from ...types.blob_storage import SeaweedSecureFilePathLocation
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import GUEST_ROLE_LEVEL
from .remote_profile import AzureRemoteProfile
from .remote_profile import RemoteProfileStash
from .stash import BlobStorageStash

BlobDepositType = Union[OnDiskBlobDeposit, SeaweedFSBlobDeposit]


@serializable()
class BlobStorageService(AbstractService):
    store: DocumentStore
    stash: BlobStorageStash
    remote_profile_stash: RemoteProfileStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = BlobStorageStash(store=store)
        self.remote_profile_stash = RemoteProfileStash(store=store)

    @service_method(path="blob_storage.get_all", name="get_all")
    def get_all_blob_storage_entries(
        self, context: AuthedServiceContext
    ) -> Union[List[BlobStorageEntry], SyftError]:
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(path="blob_storage.mount_azure", name="mount_azure")
    def mount_azure(
        self,
        context: AuthedServiceContext,
        account_name: str,
        account_key: str,
        container_name: str,
        bucket_name: str,
        use_direct_connections=True,
    ):
        # stdlib

        # TODO: fix arguments

        remote_name = f"{account_name}{container_name}"
        remote_name = "".join(ch for ch in remote_name if ch.isalnum())
        args_dict = {
            "account_name": account_name,
            "account_key": account_key,
            "container_name": container_name,
            "remote_name": remote_name,
            "bucket_name": bucket_name,
        }

        new_profile = AzureRemoteProfile(
            profile_name=remote_name,
            account_name=account_name,
            account_key=account_key,
            container_name=container_name,
        )
        res = self.remote_profile_stash.set(context.credentials, new_profile)
        if res.is_err():
            return SyftError(message=res.value)
        remote_profile = res.ok()
        seaweed_config = context.node.blob_storage_client.config
        # we cache this here such that we can use it when reading a file from azure
        # from the remote_name
        seaweed_config.remote_profiles[remote_name] = remote_profile

        # TODO: possible wrap this in try catch
        cfg = context.node.blob_store_config.client_config
        init_request = requests.post(url=cfg.mount_url, json=args_dict)  # nosec
        print(init_request.content)
        # TODO check return code
        res = context.node.blob_storage_client.connect().client.list_objects(
            Bucket=bucket_name
        )
        # stdlib
        objects = res["Contents"]
        file_sizes = [object["Size"] for object in objects]
        file_paths = [object["Key"] for object in objects]
        if use_direct_connections:
            secure_file_paths = [
                AzureSecureFilePathLocation(
                    path=file_path,
                    azure_profile_name=remote_name,
                    bucket_name=bucket_name,
                )
                for file_path in file_paths
            ]
        else:
            secure_file_paths = [
                SeaweedSecureFilePathLocation(
                    path=file_path,
                )
                for file_path in file_paths
            ]

        for sfp, file_size in zip(secure_file_paths, file_sizes):
            blob_storage_entry = BlobStorageEntry(
                location=sfp,
                uploaded_by=context.credentials,
                file_size=file_size,
                type_=BlobFileType,
                bucket_name=bucket_name,
            )
            self.stash.set(context.credentials, blob_storage_entry)

        return SyftSuccess(message="Mounting Azure Successful!")

    @service_method(
        path="blob_storage.get_files_from_bucket", name="get_files_from_bucket"
    )
    def get_files_from_bucket(self, context: AuthedServiceContext, bucket_name: str):
        result = self.stash.find_all(context.credentials, bucket_name=bucket_name)
        if result.is_err():
            return result
        bse_list = result.ok()
        # stdlib

        blob_files = []
        for bse in bse_list:
            self.stash.set(obj=bse, credentials=context.credentials)
            # We create an empty ActionObject and set its blob_storage_entry_id to bse.id
            # such that we can call reload_cache which creates
            # the BlobRetrieval (user needs permission to do this)
            # This could be a BlobRetrievalByURL that creates a BlobFile
            # and then sets it in the cache (it does not contain the data, only the BlobFile).
            # In the client, when reading the file, we will creates **another**, blobretrieval
            # object to read the actual data
            blob_file = ActionObject.empty()
            blob_file.syft_blob_storage_entry_id = bse.id
            blob_file.syft_client_verify_key = context.credentials
            blob_file.syft_node_location = context.node.id
            blob_file.reload_cache()
            blob_files.append(blob_file.syft_action_data)

        return blob_files

    @service_method(path="blob_storage.get_by_uid", name="get_by_uid")
    def get_blob_storage_entry_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[BlobStorageEntry, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(path="blob_storage.get_metadata", name="get_metadata")
    def get_blob_storage_metadata_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[BlobStorageEntry, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            blob_storage_entry = result.ok()
            return blob_storage_entry.to(BlobStorageMetadata)
        return SyftError(message=result.err())

    # TODO: replace name with `create_blob_retrieval`
    @service_method(
        path="blob_storage.read",
        name="read",
        roles=GUEST_ROLE_LEVEL,
    )
    def read(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[BlobRetrieval, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            obj: BlobStorageEntry = result.ok()
            if obj is None:
                return SyftError(message=f"No blob storage entry exists for uid: {uid}")

            with context.node.blob_storage_client.connect() as conn:
                res: BlobRetrieval = conn.read(
                    obj.location, obj.type_, bucket_name=obj.bucket_name
                )
                res.syft_blob_storage_entry_id = uid
                res.file_size = obj.file_size
                return res
        return SyftError(message=result.err())

    @service_method(
        path="blob_storage.allocate",
        name="allocate",
        roles=GUEST_ROLE_LEVEL,
    )
    def allocate(
        self, context: AuthedServiceContext, obj: CreateBlobStorageEntry
    ) -> Union[BlobDepositType, SyftError]:
        with context.node.blob_storage_client.connect() as conn:
            secure_location = conn.allocate(obj)

            if isinstance(secure_location, SyftError):
                return secure_location

            blob_storage_entry = BlobStorageEntry(
                id=obj.id,
                location=secure_location,
                type_=obj.type_,
                mimetype=obj.mimetype,
                file_size=obj.file_size,
                uploaded_by=context.credentials,
            )
            blob_deposit = conn.write(blob_storage_entry)

        result = self.stash.set(context.credentials, blob_storage_entry)
        if result.is_err():
            return SyftError(message=f"{result.err()}")
        return blob_deposit

    @service_method(
        path="blob_storage.write_to_disk",
        name="write_to_disk",
        roles=GUEST_ROLE_LEVEL,
    )
    def write_to_disk(
        self, context: AuthedServiceContext, uid: UID, data: bytes
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(
            credentials=context.credentials,
            uid=uid,
        )
        if result.is_err():
            return SyftError(message=f"{result.err()}")

        obj: Optional[BlobStorageEntry] = result.ok()

        if obj is None:
            return SyftError(message=f"No blob storage entry exists for uid: {uid}")

        try:
            Path(obj.location.path).write_bytes(data)
            return SyftSuccess(message="File successfully saved.")
        except Exception as e:
            return SyftError(message=f"Failed to write object to disk: {e}")

    @service_method(
        path="blob_storage.mark_write_complete",
        name="mark_write_complete",
        roles=GUEST_ROLE_LEVEL,
    )
    def mark_write_complete(
        self,
        context: AuthedServiceContext,
        uid: UID,
        etags: List,
        no_lines: Optional[int] = 0,
    ) -> Union[SyftError, SyftSuccess]:
        result = self.stash.get_by_uid(
            credentials=context.credentials,
            uid=uid,
        )
        if result.is_err():
            return SyftError(message=f"{result.err()}")

        obj: Optional[BlobStorageEntry] = result.ok()

        if obj is None:
            return SyftError(message=f"No blob storage entry exists for uid: {uid}")

        obj.no_lines = no_lines
        result = self.stash.update(
            credentials=context.credentials,
            obj=obj,
        )
        if result.is_err():
            return SyftError(message=f"{result.err()}")

        with context.node.blob_storage_client.connect() as conn:
            result = conn.complete_multipart_upload(obj, etags)

        return result

    @service_method(path="blob_storage.delete", name="delete")
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            obj = result.ok()

            if obj is None:
                return SyftError(message=f"No blob storage entry exists for uid: {uid}")
            try:
                with context.node.blob_storage_client.connect() as conn:
                    file_unlinked_result = conn.delete(obj.location)
            except Exception as e:
                return SyftError(message=f"Failed to delete file: {e}")

            if isinstance(file_unlinked_result, SyftError):
                return file_unlinked_result
            blob_storage_entry_deleted = self.stash.delete(
                context.credentials, UIDPartitionKey.with_obj(uid), has_permission=True
            )
            if blob_storage_entry_deleted.is_ok():
                return file_unlinked_result

        return SyftError(message=result.err())


TYPE_TO_SERVICE[BlobStorageEntry] = BlobStorageEntry



# stdlib
import inspect
from inspect import Parameter
from inspect import Signature
from inspect import _ParameterKind
from inspect import _signature_fromstr
import re
from typing import Callable
from typing import Optional

# relative
from .deserialize import _deserialize
from .recursive import recursive_serde_register
from .serialize import _serialize

recursive_serde_register(_ParameterKind)


recursive_serde_register(
    Parameter, serialize_attrs=["_annotation", "_name", "_kind", "_default"]
)


# def serialize_parameter(obj: Parameter) -> bytes:
#     # 🟡 TODO 3: Solve issue of Signature Parameter types being converted to String depending
#                  on import path
#     # currently types arent always being sent correctly maybe due to the path?
#     # instead we can send the fqn and recover it in the type checker on the client side
#     annotation = obj.annotation
#     if not isinstance(annotation, str):
#         annotation = f"{annotation.__module__}.{annotation.__name__}"
#     obj_dict = {
#         "name": obj.name,
#         "kind": obj.kind,
#         "default": obj.default,
#         "annotation": annotation,
#     }
#     return _serialize(obj_dict, to_bytes=True)


# def deserialize_parameter(blob: bytes) -> Parameter:
#     obj_dict = _deserialize(blob, from_bytes=True)
#     return Parameter(**obj_dict)


# recursive_serde_register(Parameter, serialize_parameter, deserialize_parameter)


def serialize_signature(obj: Signature) -> bytes:
    parameters = list(dict(obj.parameters).values())
    return_annotation = obj.return_annotation
    obj_dict = {"parameters": parameters, "return_annotation": return_annotation}
    return _serialize(obj_dict, to_bytes=True)


def deserialize_signature(blob: bytes) -> Signature:
    obj_dict = _deserialize(blob, from_bytes=True)
    return Signature(**obj_dict)


recursive_serde_register(Signature, serialize_signature, deserialize_signature)


def signature_remove_self(signature: Signature) -> Signature:
    params = dict(signature.parameters)
    params.pop("self", None)
    return Signature(
        list(params.values()), return_annotation=signature.return_annotation
    )


def signature_remove_context(signature: Signature) -> Signature:
    params = dict(signature.parameters)
    params.pop("context", None)
    return Signature(
        list(params.values()), return_annotation=signature.return_annotation
    )


def get_str_signature_from_docstring(doc: str, callable_name: str) -> Optional[str]:
    if not doc or callable_name not in doc:
        return None
    else:
        # gets part of the docstring that contains the signature
        doc = re.sub(r"\s", "", doc.split("\n\n")[0])
        # gets signature
        signature_search_result = re.search(rf"{callable_name}\((.+)\)", doc)
        if signature_search_result:
            signature_str = signature_search_result.group(1)
            # decomposing "[]" optional  params
            params = re.findall(r"\[(.+?)\]", signature_str)

            if params:
                # for instance in:
                # numpy.matmul(x1, x2, /, out=None, *, casting='same_kind',
                # order='K', dtype=None, subok=True[, signature, extobj, axes, axis])
                # we have [, signature, extobj, axes, axis] , which we replace as normal params
                for param in params[:-1]:
                    signature_str = signature_str.replace(f"[{param}]", param)

                if re.search(rf"(?<={params[-1]})\],", signature_str):
                    signature_str = signature_str.replace(
                        f"[{params[-1]}],", params[-1]
                    )
                else:
                    signature_str = signature_str.replace(
                        f"[{params[-1]}]",
                        f', {",".join([f"{param}=None" for param in params[-1].split(",") if param])}',
                    )

            # removes "/" and "*"
            signature_str = re.sub(r",(\/|\*)", "", signature_str)
            # removes dtypes
            signature_str = re.sub(r"dtype=(\w+),", "dtype=None,", signature_str)
            return f"{callable_name}({signature_str})"
        else:
            return None


def generate_signature(_callable: Callable) -> inspect.Signature:
    name, doc = _callable.__name__, _callable.__doc__
    # returning predefined signature if in signature registry
    # name_in_registry = name in function_signatures_registry.keys()
    text_signature = get_str_signature_from_docstring(doc, name)
    # TODO safe handling if function signature can not be generated
    return _signature_fromstr(inspect.Signature, _callable, text_signature, True)


def get_signature(_callable: Callable) -> inspect.Signature:
    try:
        res = inspect.signature(_callable)
        if res is None:
            raise ValueError("")
        else:
            return res
    except Exception:
        return generate_signature(_callable)



__version__ = "0.8.4-beta.13"

# stdlib
import pathlib
from pathlib import Path
import sys
from typing import Any
from typing import Callable

# relative
from . import gevent_patch  # noqa: F401
from .abstract_node import NodeType  # noqa: F401
from .client.client import connect  # noqa: F401
from .client.client import login  # noqa: F401
from .client.client import login_as_guest  # noqa: F401
from .client.client import register  # noqa: F401
from .client.deploy import Orchestra  # noqa: F401
from .client.domain_client import DomainClient  # noqa: F401
from .client.gateway_client import GatewayClient  # noqa: F401
from .client.registry import DomainRegistry  # noqa: F401
from .client.registry import EnclaveRegistry  # noqa: F401
from .client.registry import NetworkRegistry  # noqa: F401
from .client.search import Search  # noqa: F401
from .client.search import SearchResults  # noqa: F401
from .client.user_settings import UserSettings  # noqa: F401
from .client.user_settings import settings  # noqa: F401
from .custom_worker.config import DockerWorkerConfig  # noqa: F401
from .external import OBLV  # noqa: F401
from .external import enable_external_lib  # noqa: F401
from .node.credentials import SyftSigningKey  # noqa: F401
from .node.domain import Domain  # noqa: F401
from .node.enclave import Enclave  # noqa: F401
from .node.gateway import Gateway  # noqa: F401
from .node.server import serve_node  # noqa: F401
from .node.server import serve_node as bind_worker  # noqa: F401
from .node.worker import Worker  # noqa: F401
from .protocol.data_protocol import bump_protocol_version  # noqa: F401
from .protocol.data_protocol import check_or_stage_protocol  # noqa: F401
from .protocol.data_protocol import get_data_protocol  # noqa: F401
from .protocol.data_protocol import stage_protocol_changes  # noqa: F401
from .serde import NOTHING  # noqa: F401
from .serde.deserialize import _deserialize as deserialize  # noqa: F401
from .serde.serializable import serializable  # noqa: F401
from .serde.serialize import _serialize as serialize  # noqa: F401
from .service.action.action_data_empty import ActionDataEmpty  # noqa: F401
from .service.action.action_object import ActionObject  # noqa: F401
from .service.action.plan import Plan  # noqa: F401
from .service.action.plan import planify  # noqa: F401
from .service.code.user_code import UserCodeStatus  # noqa: F401; noqa: F401
from .service.code.user_code import syft_function  # noqa: F401; noqa: F401
from .service.code.user_code import syft_function_single_use  # noqa: F401; noqa: F401
from .service.data_subject import DataSubjectCreate as DataSubject  # noqa: F401
from .service.dataset.dataset import Contributor  # noqa: F401
from .service.dataset.dataset import CreateAsset as Asset  # noqa: F401
from .service.dataset.dataset import CreateDataset as Dataset  # noqa: F401
from .service.notification.notifications import NotificationStatus  # noqa: F401
from .service.policy.policy import CustomInputPolicy  # noqa: F401
from .service.policy.policy import CustomOutputPolicy  # noqa: F401
from .service.policy.policy import ExactMatch  # noqa: F401
from .service.policy.policy import SingleExecutionExactOutput  # noqa: F401
from .service.policy.policy import UserInputPolicy  # noqa: F401
from .service.policy.policy import UserOutputPolicy  # noqa: F401
from .service.project.project import ProjectSubmit as Project  # noqa: F401
from .service.request.request import SubmitRequest as Request  # noqa: F401
from .service.response import SyftError  # noqa: F401
from .service.response import SyftNotReady  # noqa: F401
from .service.response import SyftSuccess  # noqa: F401
from .service.user.roles import Roles as roles  # noqa: F401
from .service.user.user_service import UserService  # noqa: F401
from .types.twin_object import TwinObject  # noqa: F401
from .types.uid import UID  # noqa: F401
from .util import filterwarnings  # noqa: F401
from .util import jax_settings  # noqa: F401
from .util import logger  # noqa: F401
from .util import options  # noqa: F401
from .util.autoreload import disable_autoreload  # noqa: F401
from .util.autoreload import enable_autoreload  # noqa: F401
from .util.telemetry import instrument  # noqa: F401
from .util.util import autocache  # noqa: F401
from .util.util import get_root_data_path  # noqa: F401
from .util.version_compare import make_requires

LATEST_STABLE_SYFT = "0.8.3"
requires = make_requires(LATEST_STABLE_SYFT, __version__)


# SYFT_PATH = path = os.path.abspath(a_module.__file__)
SYFT_PATH = pathlib.Path(__file__).parent.resolve()

sys.path.append(str(Path(__file__)))

logger.start()

try:
    get_ipython()  # noqa: F821
    # TODO: add back later or auto detect
    # display(
    #     Markdown(
    #         "\nWarning: syft is imported in light mode by default. \
    #     \nTo switch to dark mode, please run `sy.options.color_theme = 'dark'`"
    #     )
    # )
except:  # noqa: E722
    pass  # nosec

# For server-side, to enable by environment variable
if OBLV:
    enable_external_lib("oblv")


def module_property(func: Any) -> Callable:
    """Decorator to turn module functions into properties.
    Function names must be prefixed with an underscore."""
    module = sys.modules[func.__module__]

    def base_getattr(name: str) -> None:
        raise AttributeError(f"module {module.__name__!r} has no attribute {name!r}")

    old_getattr = getattr(module, "__getattr__", base_getattr)

    def new_getattr(name: str) -> Any:
        if f"_{name}" == func.__name__:
            return func()
        else:
            return old_getattr(name)

    module.__getattr__ = new_getattr  # type: ignore
    return func


@module_property
def _gateways() -> NetworkRegistry:
    return NetworkRegistry()


@module_property
def _enclaves() -> EnclaveRegistry:
    return EnclaveRegistry()


@module_property
def _domains() -> DomainRegistry:
    return DomainRegistry()


@module_property
def _settings() -> UserSettings:
    return settings


@module_property
def _orchestra() -> Orchestra:
    return Orchestra


def search(name: str) -> SearchResults:
    return Search(_domains()).search(name=name)



# relative
from ..abstract_node import NodeType
from ..serde.serializable import serializable
from .node import Node


@serializable()
class Gateway(Node):
    def post_init(self) -> None:
        self.node_type = NodeType.GATEWAY
        super().post_init()



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import pydantic

# relative
from ..serde.serializable import serializable
from ..types.base import SyftBaseModel

SIGNING_KEY_FOR = "SigningKey for"


@serializable()
class SyftVerifyKey(SyftBaseModel):
    verify_key: VerifyKey

    def __init__(self, verify_key: Union[str, VerifyKey]):
        if isinstance(verify_key, str):
            verify_key = VerifyKey(bytes.fromhex(verify_key))
        super().__init__(verify_key=verify_key)

    def __str__(self) -> str:
        return self.verify_key.encode(encoder=HexEncoder).decode("utf-8")

    @staticmethod
    def from_string(key_str: str) -> SyftVerifyKey:
        return SyftVerifyKey(verify_key=VerifyKey(bytes.fromhex(key_str)))

    @property
    def verify(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SyftVerifyKey):
            return False
        return self.verify_key == other.verify_key

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(self.verify_key)


@serializable()
class SyftSigningKey(SyftBaseModel):
    signing_key: SigningKey

    @pydantic.validator("signing_key", pre=True, always=True)
    def make_signing_key(cls, v: Union[str, SigningKey]) -> SigningKey:
        return SigningKey(bytes.fromhex(v)) if isinstance(v, str) else v

    @property
    def verify_key(self) -> SyftVerifyKey:
        return SyftVerifyKey(verify_key=self.signing_key.verify_key)

    def __str__(self) -> str:
        return self.signing_key.encode(encoder=HexEncoder).decode("utf-8")

    @staticmethod
    def generate() -> SyftSigningKey:
        return SyftSigningKey(signing_key=SigningKey.generate())

    @staticmethod
    def from_string(key_str: str) -> SyftSigningKey:
        return SyftSigningKey(signing_key=SigningKey(bytes.fromhex(key_str)))

    def __repr__(self) -> str:
        return f"<{SIGNING_KEY_FOR}: {self.verify}>"

    @property
    def verify(self) -> str:
        return str(self.verify_key)

    def __hash__(self) -> int:
        return hash(self.signing_key)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SyftSigningKey):
            return False
        return self.signing_key == other.signing_key


SyftCredentials = Union[SyftVerifyKey, SyftSigningKey]


@serializable()
class UserLoginCredentials(SyftBaseModel):
    email: str
    password: str



# stdlib
import contextlib
import io
import os.path
from pathlib import Path
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Tuple

# third party
import docker
from docker.models.images import Image

# relative
from .config import CustomWorkerConfig
from .config import DockerWorkerConfig
from .config import WorkerConfig


class CustomWorkerBuilder:
    TYPE_CPU = "cpu"
    TYPE_GPU = "gpu"

    TEMPLATE_DIR_PROD = os.path.expandvars("$APPDIR/grid/")
    TEMPLATE_DIR_DEV = "../../../../../grid/backend/"

    CUSTOM_IMAGE_PREFIX = "custom-worker"

    BUILD_MAX_WAIT = 30 * 60

    def build_image(
        self,
        config: WorkerConfig,
        tag: str = None,
        **kwargs: Any,
    ) -> Tuple[Image, Iterable[str]]:
        """
        Builds a Docker image from the given configuration.
        Args:
            config (WorkerConfig): The configuration for building the Docker image.
            tag (str): The tag to use for the image.
        """

        if isinstance(config, DockerWorkerConfig):
            return self._build_dockerfile(config, tag, **kwargs)
        elif isinstance(config, CustomWorkerConfig):
            return self._build_template(config, **kwargs)
        else:
            raise TypeError("Unknown worker config type")

    def push_image(self, tag: str, **kwargs: Any) -> str:
        """
        Pushes a Docker image to the given repo.
        Args:
            repo (str): The repo to push the image to.
            tag (str): The tag to use for the image.
        """

        return self._push_image(tag, **kwargs)

    def _build_dockerfile(self, config: DockerWorkerConfig, tag: str, **kwargs):
        # convert string to file-like object
        file_obj = io.BytesIO(config.dockerfile.encode("utf-8"))
        return self._build_image(fileobj=file_obj, tag=tag, **kwargs)

    def _build_template(self, config: CustomWorkerConfig, **kwargs: Any):
        # Builds a Docker pre-made CPU/GPU image template using a CustomWorkerConfig
        # remove once GPU is supported
        if config.build.gpu:
            raise Exception("GPU custom worker is not supported yet")

        type = self.TYPE_GPU if config.build.gpu else self.TYPE_CPU

        contextdir, dockerfile = self._find_template_dir(type)

        imgtag = config.get_signature()[:8]

        build_args = {
            "PYTHON_VERSION": config.build.python_version,
            "SYSTEM_PACKAGES": config.build.merged_system_pkgs(),
            "PIP_PACKAGES": config.build.merged_python_pkgs(),
            "CUSTOM_CMD": config.build.merged_custom_cmds(),
        }

        return self._build_image(
            tag=f"{self.CUSTOM_IMAGE_PREFIX}-{type}:{imgtag}",
            path=str(contextdir),
            dockerfile=dockerfile,
            buildargs=build_args,
        )

    def _build_image(self, tag: str, **build_opts) -> Tuple[Image, Iterable]:
        # Core docker build call. Func signature should match with Docker SDK's BuildApiMixin
        with contextlib.closing(docker.from_env()) as client:
            image_result = client.images.build(
                tag=tag,
                timeout=self.BUILD_MAX_WAIT,
                **build_opts,
            )
            return image_result

    def _push_image(
        self,
        tag: str,
        registry_url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> str:
        with contextlib.closing(docker.from_env()) as client:
            if registry_url and username and password:
                client.login(
                    username=username, password=password, registry=registry_url
                )

            result = client.images.push(repository=tag)
            return result

    def _find_template_dir(self, type: str) -> Tuple[Path, str]:
        """
        Find the Worker Dockerfile and it's context path
        - PROD will be in `$APPDIR/grid/`
        - DEV will be in `packages/grid/backend`
        - In both the cases context dir does not matter (unless we're calling COPY)

        Args:
            type (str): The type of worker.
        Returns:
            Path: The path to the Dockerfile.
        """
        filename = f"worker_{type}.dockerfile"
        lookup_paths = [
            Path(self.TEMPLATE_DIR_PROD, filename).resolve(),
            Path(__file__, self.TEMPLATE_DIR_DEV, filename).resolve(),
        ]
        for path in lookup_paths:
            if path.exists():
                return path.parent, filename

        raise FileNotFoundError(f"Cannot find the {filename}")



# stdlib
import logging
import os
import sys
from typing import Any
from typing import Callable
from typing import NoReturn
from typing import TextIO
from typing import Union

# third party
from loguru import logger

LOG_FORMAT = "[{time}][{level}][{module}]][{process.id}] {message}"

logger.remove()
DEFAULT_SINK = "syft_{time}.log"


def remove() -> None:
    logger.remove()


def add(
    sink: Union[None, str, os.PathLike, TextIO, logging.Handler] = None,
    level: str = "ERROR",
) -> None:
    sink = DEFAULT_SINK if sink is None else sink
    try:
        logger.add(
            sink=sink,
            format=LOG_FORMAT,
            enqueue=True,
            colorize=False,
            diagnose=True,
            backtrace=True,
            rotation="10 MB",
            retention="1 day",
            level=level,
        )
    except BaseException:
        logger.add(
            sink=sink,
            format=LOG_FORMAT,
            colorize=False,
            diagnose=True,
            backtrace=True,
            level=level,
        )


def start() -> None:
    add(sink=sys.stderr, level="CRITICAL")


def stop() -> None:
    logger.stop()


def traceback_and_raise(e: Any, verbose: bool = False) -> NoReturn:
    try:
        if verbose:
            logger.opt(lazy=True).exception(e)
        else:
            logger.opt(lazy=True).critical(e)
    except BaseException as ex:
        logger.debug("failed to print exception", ex)
    if not issubclass(type(e), Exception):
        e = Exception(e)
    raise e


def create_log_and_print_function(level: str) -> Callable:
    def log_and_print(*args: Any, **kwargs: Any) -> None:
        try:
            method = getattr(logger.opt(lazy=True), level, None)
            if "print" in kwargs and kwargs["print"] is True:
                del kwargs["print"]
                print(*args, **kwargs)
                if "end" in kwargs:
                    # clean up extra end for printinga
                    del kwargs["end"]

            if method is not None:
                method(*args, **kwargs)
            else:
                raise Exception(f"no method {level} on logger")
        except BaseException as e:
            msg = f"failed to log exception. {e}"
            try:
                logger.debug(msg)

            except Exception as e:
                print(f"{msg}. {e}")

    return log_and_print


def traceback(*args: Any, **kwargs: Any) -> None:
    # caller = inspect.getframeinfo(inspect.stack()[1][0])
    # print(f"traceback:{caller.filename}:{caller.function}:{caller.lineno}")
    return create_log_and_print_function(level="exception")(*args, **kwargs)


def critical(*args: Any, **kwargs: Any) -> None:
    # caller = inspect.getframeinfo(inspect.stack()[1][0])
    # print(f"critical:{caller.filename}:{caller.function}:{caller.lineno}:{args}")
    return create_log_and_print_function(level="critical")(*args, **kwargs)


def error(*args: Any, **kwargs: Any) -> None:
    # caller = inspect.getframeinfo(inspect.stack()[1][0])
    # print(f"error:{caller.filename}:{caller.function}:{caller.lineno}")
    return create_log_and_print_function(level="error")(*args, **kwargs)


def warning(*args: Any, **kwargs: Any) -> None:
    return create_log_and_print_function(level="warning")(*args, **kwargs)


def info(*args: Any, **kwargs: Any) -> None:
    return create_log_and_print_function(level="info")(*args, **kwargs)


def debug(*args: Any) -> None:
    debug_msg = " ".join([str(a) for a in args])
    return logger.debug(debug_msg)


def _debug(*args: Any, **kwargs: Any) -> None:
    return create_log_and_print_function(level="debug")(*args, **kwargs)


def trace(*args: Any, **kwargs: Any) -> None:
    return create_log_and_print_function(level="trace")(*args, **kwargs)



# stdlib
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import pydantic

# relative
from ...custom_worker.config import CustomWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.dicttuple import DictTuple
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..request.request import Change
from ..request.request import CreateCustomImageChange
from ..request.request import CreateCustomWorkerPoolChange
from ..request.request import SubmitRequest
from ..request.request_service import RequestService
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .image_identifier import SyftWorkerImageIdentifier
from .utils import DEFAULT_WORKER_POOL_NAME
from .utils import run_containers
from .utils import run_workers_in_threads
from .worker_image import SyftWorkerImage
from .worker_image_stash import SyftWorkerImageStash
from .worker_pool import ContainerSpawnStatus
from .worker_pool import WorkerOrchestrationType
from .worker_pool import WorkerPool
from .worker_pool_stash import SyftWorkerPoolStash
from .worker_service import WorkerService
from .worker_stash import WorkerStash


@serializable()
class SyftWorkerPoolService(AbstractService):
    store: DocumentStore
    stash: SyftWorkerPoolStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SyftWorkerPoolStash(store=store)
        self.image_stash = SyftWorkerImageStash(store=store)

    @service_method(
        path="worker_pool.launch",
        name="launch",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def launch(
        self,
        context: AuthedServiceContext,
        name: str,
        image_uid: Optional[UID],
        num_workers: int,
        reg_username: Optional[str] = None,
        reg_password: Optional[str] = None,
    ) -> Union[List[ContainerSpawnStatus], SyftError]:
        """Creates a pool of workers from the given SyftWorkerImage.

        - Retrieves the image for the given UID
        - Use docker to launch containers for given image
        - For each successful container instantiation create a SyftWorker object
        - Creates a SyftWorkerPool object

        Args:
            context (AuthedServiceContext): context passed to the service
            name (str): name of the pool
            image_id (UID): UID of the SyftWorkerImage against which the pool should be created
            num_workers (int): the number of SyftWorker that needs to be created in the pool
        """

        result = self.stash.get_by_name(context.credentials, pool_name=name)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        if result.ok() is not None:
            return SyftError(message=f"Worker Pool with name: {name} already exists !!")

        # If image uid is not passed, then use the default worker image
        # to create the worker pool
        if image_uid is None:
            result = self.stash.get_by_name(
                context.credentials, pool_name=DEFAULT_WORKER_POOL_NAME
            )
            default_worker_pool = result.ok()
            image_uid = default_worker_pool.image_id

        # Get the image object for the given image id
        result = self.image_stash.get_by_uid(
            credentials=context.credentials, uid=image_uid
        )
        if result.is_err():
            return SyftError(
                message=f"Failed to retrieve Worker Image with id: {image_uid}. Error: {result.err()}"
            )

        worker_image: SyftWorkerImage = result.ok()

        worker_service: WorkerService = context.node.get_service("WorkerService")
        worker_stash = worker_service.stash

        # Create worker pool from given image, with the given worker pool
        # and with the desired number of workers
        worker_list, container_statuses = _create_workers_in_pool(
            context=context,
            pool_name=name,
            existing_worker_cnt=0,
            worker_cnt=num_workers,
            worker_image=worker_image,
            worker_stash=worker_stash,
            reg_username=reg_username,
            reg_password=reg_password,
        )

        # Update the Database with the pool information
        worker_pool = WorkerPool(
            name=name,
            max_count=num_workers,
            image_id=worker_image.id,
            worker_list=worker_list,
            syft_node_location=context.node.id,
            syft_client_verify_key=context.credentials,
        )
        result = self.stash.set(credentials=context.credentials, obj=worker_pool)

        if result.is_err():
            return SyftError(message=f"Failed to save Worker Pool: {result.err()}")

        return container_statuses

    @service_method(
        path="worker_pool.create_pool_request",
        name="pool_creation_request",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def create_pool_request(
        self,
        context: AuthedServiceContext,
        pool_name: str,
        num_workers: int,
        image_uid: UID,
        reason: Optional[str] = "",
    ) -> Union[SyftError, SyftSuccess]:
        """
        Create a request to launch the worker pool based on a built image.

        Args:
            context (AuthedServiceContext): The authenticated service context.
            pool_name (str): The name of the worker pool.
            num_workers (int): The number of workers in the pool.
            image_uid (Optional[UID]): The UID of the built image.
            reason (Optional[str], optional): The reason for creating the
                worker pool. Defaults to "".
        """

        # Check if image exists for the given image id
        search_result = self.image_stash.get_by_uid(
            credentials=context.credentials, uid=image_uid
        )

        if search_result.is_err():
            return SyftError(message=str(search_result.err()))

        worker_image: Optional[SyftWorkerImage] = search_result.ok()

        # Raise error if worker image doesn't exists
        if worker_image is None:
            return SyftError(
                message=f"No image exists for given image uid : {image_uid}"
            )

        # Check if pool already exists for the given pool name
        result = self.stash.get_by_name(context.credentials, pool_name=pool_name)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        worker_pool = result.ok()

        if worker_pool is not None:
            return SyftError(
                message=f"Worker pool already exists for given pool name: {pool_name}"
            )

        # If no worker pool exists for given pool name
        # and image exists for given image uid, then create a change
        # request object to create the pool with the desired number of workers
        create_worker_pool_change = CreateCustomWorkerPoolChange(
            pool_name=pool_name,
            num_workers=num_workers,
            image_uid=image_uid,
        )

        changes: List[Change] = [create_worker_pool_change]

        # Create a the request object with the changes and submit it
        # for approval.
        request = SubmitRequest(changes=changes)
        method = context.node.get_service_method(RequestService.submit)
        result = method(context=context, request=request, reason=reason)

        return result

    @service_method(
        path="worker_pool.create_image_and_pool_request",
        name="create_image_and_pool_request",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def create_image_and_pool_request(
        self,
        context: AuthedServiceContext,
        pool_name: str,
        num_workers: int,
        tag: str,
        config: WorkerConfig,
        reason: Optional[str] = "",
    ) -> Union[SyftError, SyftSuccess]:
        """
        Create a request to launch the worker pool based on a built image.

        Args:
            context (AuthedServiceContext): The authenticated service context.
            pool_name (str): The name of the worker pool.
            num_workers (int): The number of workers in the pool.
            config: (WorkerConfig): Config of the image to be built.
            tag (str): human-readable manifest identifier that is typically a specific version or variant of an image
            reason (Optional[str], optional): The reason for creating the worker image and pool. Defaults to "".
        """

        if isinstance(config, CustomWorkerConfig):
            return SyftError(message="We only support DockerWorkerConfig.")

        # Check if an image already exists for given docker config
        search_result = self.image_stash.get_by_docker_config(
            credentials=context.credentials, config=config
        )

        if search_result.is_err():
            return SyftError(message=str(search_result.err()))

        worker_image: Optional[SyftWorkerImage] = search_result.ok()

        if worker_image is not None:
            return SyftError(
                message="Image already exists for given config. \
                    Please use `worker_pool.create_pool_request` to request pool creation."
            )

        # Validate Image Tag
        try:
            image_identifier = SyftWorkerImageIdentifier.from_str(tag=tag)
        except pydantic.ValidationError as e:
            return SyftError(message=f"Failed to create tag: {e}")

        # create a list of Change objects and submit a
        # request for these changes for approval
        changes: List[Change] = []

        # Add create custom image change
        # If this change is approved, then build an image using the config
        create_custom_image_change = CreateCustomImageChange(
            config=config,
            tag=image_identifier.full_name_with_tag,
        )

        # Check if a pool already exists for given pool name
        result = self.stash.get_by_name(context.credentials, pool_name=pool_name)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        # Raise an error if worker pool already exists for the given worker pool name
        if result.ok() is not None:
            return SyftError(
                message=f"Worker Pool with name: {pool_name} already "
                f"exists. Please choose another name!"
            )

        # Add create worker pool change
        # If change is approved then worker pool is created and
        # the desired number of workers are added to the pool
        create_worker_pool_change = CreateCustomWorkerPoolChange(
            pool_name=pool_name,
            num_workers=num_workers,
            config=config,
        )
        changes += [create_custom_image_change, create_worker_pool_change]

        # Create a request object and submit a request for approval
        request = SubmitRequest(changes=changes)
        method = context.node.get_service_method(RequestService.submit)
        result = method(context=context, request=request, reason=reason)

        return result

    @service_method(
        path="worker_pool.get_all",
        name="get_all",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all(
        self, context: AuthedServiceContext
    ) -> Union[DictTuple[str, WorkerPool], SyftError]:
        # TODO: During get_all, we should dynamically make a call to docker to get the status of the containers
        # and update the status of the workers in the pool.
        result = self.stash.get_all(credentials=context.credentials)
        if result.is_err():
            return SyftError(message=f"{result.err()}")
        worker_pools: List[WorkerPool] = result.ok()

        res: List[Tuple] = []
        for pool in worker_pools:
            res.append((pool.name, pool))
        return DictTuple(res)

    @service_method(
        path="worker_pool.add_workers",
        name="add_workers",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def add_workers(
        self,
        context: AuthedServiceContext,
        number: int,
        pool_id: Optional[UID] = None,
        pool_name: Optional[str] = None,
    ) -> Union[List[ContainerSpawnStatus], SyftError]:
        """Add workers to existing worker pool.

        Worker pool is fetched either using the unique pool id or pool name.

        Args:
            context (AuthedServiceContext): _description_
            number (int): number of workers to add
            pool_id (Optional[UID], optional): Unique UID of the pool. Defaults to None.
            pool_name (Optional[str], optional): Unique name of the pool. Defaults to None.

        Returns:
            Union[List[ContainerSpawnStatus], SyftError]: List of spawned workers with their status and error if any.
        """

        # Extract pool using either using pool id or pool name
        if pool_id:
            result = self.stash.get_by_uid(credentials=context.credentials, uid=pool_id)
        elif pool_name:
            result = self.stash.get_by_name(
                credentials=context.credentials,
                pool_name=pool_name,
            )

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        worker_pool = result.ok()

        existing_worker_cnt = len(worker_pool.worker_list)

        result = self.image_stash.get_by_uid(
            credentials=context.credentials,
            uid=worker_pool.image_id,
        )

        if result.is_err():
            return SyftError(
                message=f"Failed to retrieve image for worker pool: {worker_pool.name}"
            )

        worker_image: SyftWorkerImage = result.ok()

        worker_service: WorkerService = context.node.get_service("WorkerService")
        worker_stash = worker_service.stash

        # Add workers to given pool from the given image
        worker_list, container_statuses = _create_workers_in_pool(
            context=context,
            pool_name=worker_pool.name,
            existing_worker_cnt=existing_worker_cnt,
            worker_cnt=number,
            worker_image=worker_image,
            worker_stash=worker_stash,
        )

        worker_pool.worker_list += worker_list
        worker_pool.max_count = existing_worker_cnt + number

        update_result = self.stash.update(
            credentials=context.credentials, obj=worker_pool
        )
        if update_result.is_err():
            return SyftError(
                message=f"Failed update worker pool: {worker_pool.name} with err: {result.err()}"
            )

        return container_statuses

    @service_method(
        path="worker_pool.filter_by_image_id",
        name="filter_by_image_id",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def filter_by_image_id(
        self, context: AuthedServiceContext, image_uid: UID
    ) -> Union[List[WorkerPool], SyftError]:
        result = self.stash.get_by_image_uid(context.credentials, image_uid)

        if result.is_err():
            return SyftError(message=f"Failed to get worker pool for uid: {image_uid}")

        return result.ok()

    def _get_worker_pool(
        self,
        context: AuthedServiceContext,
        worker_pool_id: UID,
    ) -> Union[WorkerPool, SyftError]:
        result = self.stash.get_by_uid(
            credentials=context.credentials, uid=worker_pool_id
        )

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        worker_pool = result.ok()

        return (
            SyftError(message=f"worker pool with id {worker_pool_id} does not exist")
            if worker_pool is None
            else worker_pool
        )


def _create_workers_in_pool(
    context: AuthedServiceContext,
    pool_name: str,
    existing_worker_cnt: int,
    worker_cnt: int,
    worker_image: SyftWorkerImage,
    worker_stash: WorkerStash,
    reg_username: Optional[str] = None,
    reg_password: Optional[str] = None,
) -> Tuple[List[LinkedObject], List[ContainerSpawnStatus]]:
    queue_port = context.node.queue_config.client_config.queue_port

    # Check if workers needs to be run in memory or as containers
    start_workers_in_memory = context.node.in_memory_workers

    if start_workers_in_memory:
        # Run in-memory workers in threads
        container_statuses: List[ContainerSpawnStatus] = run_workers_in_threads(
            node=context.node,
            pool_name=pool_name,
            start_idx=existing_worker_cnt,
            number=worker_cnt + existing_worker_cnt,
        )
    else:
        container_statuses: List[ContainerSpawnStatus] = run_containers(
            pool_name=pool_name,
            worker_image=worker_image,
            start_idx=existing_worker_cnt,
            number=worker_cnt + existing_worker_cnt,
            orchestration=WorkerOrchestrationType.DOCKER,
            queue_port=queue_port,
            dev_mode=context.node.dev_mode,
            username=reg_username,
            password=reg_password,
            registry_url=worker_image.image_identifier.registry_host,
        )

    linked_worker_list = []

    for container_status in container_statuses:
        worker = container_status.worker
        if worker is None:
            continue
        result = worker_stash.set(
            credentials=context.credentials,
            obj=worker,
        )

        if result.is_ok():
            worker_obj = LinkedObject.from_obj(
                obj=result.ok(),
                service_type=WorkerService,
                node_uid=context.node.id,
            )
            linked_worker_list.append(worker_obj)
        else:
            container_status.error = result.err()

    return linked_worker_list, container_statuses


TYPE_TO_SERVICE[WorkerPool] = SyftWorkerPoolService
SERVICE_TO_TYPES[SyftWorkerPoolService] = WorkerPool



# future
from __future__ import annotations

# stdlib
from typing import Optional
from typing import TYPE_CHECKING

# relative
from ..abstract_node import NodeSideType
from ..client.api import APIRegistry
from ..img.base64 import base64read
from ..serde.serializable import serializable
from ..service.network.routes import NodeRouteType
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject
from ..types.uid import UID
from ..util.fonts import fonts_css
from .api import APIModule
from .client import SyftClient
from .client import login
from .client import login_as_guest

if TYPE_CHECKING:
    # relative
    from ..service.code.user_code import SubmitUserCode


@serializable()
class EnclaveMetadata(SyftObject):
    __canonical_name__ = "EnclaveMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    route: NodeRouteType


@serializable()
class EnclaveClient(SyftClient):
    # TODO: add widget repr for enclave client

    __api_patched = False

    @property
    def code(self) -> Optional[APIModule]:
        if self.api.has_service("code"):
            res = self.api.services.code
            # the order is important here
            # its also important that patching only happens once
            if not self.__api_patched:
                self._request_code_execution = res.request_code_execution
                self.__api_patched = True
            res.request_code_execution = self.request_code_execution
            return res
        return None

    @property
    def requests(self) -> Optional[APIModule]:
        if self.api.has_service("request"):
            return self.api.services.request
        return None

    def connect_to_gateway(
        self,
        via_client: Optional[SyftClient] = None,
        url: Optional[str] = None,
        port: Optional[int] = None,
        handle: Optional[NodeHandle] = None,  # noqa: F821
        email: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        if via_client is not None:
            client = via_client
        elif handle is not None:
            client = handle.client
        else:
            client = (
                login_as_guest(url=url, port=port)
                if email is None
                else login(url=url, port=port, email=email, password=password)
            )
            if isinstance(client, SyftError):
                return client

        res = self.exchange_route(client)
        if isinstance(res, SyftSuccess):
            return SyftSuccess(
                message=f"Connected {self.metadata.node_type} to {client.name} gateway"
            )
        return res

    def get_enclave_metadata(self) -> EnclaveMetadata:
        return EnclaveMetadata(route=self.connection.route)

    def request_code_execution(self, code: SubmitUserCode):
        # relative
        from ..service.code.user_code_service import SubmitUserCode

        if not isinstance(code, SubmitUserCode):
            raise Exception(
                f"The input code should be of type: {SubmitUserCode} got:{type(code)}"
            )

        enclave_metadata = self.get_enclave_metadata()

        code_id = UID()
        code.id = code_id
        code.enclave_metadata = enclave_metadata

        apis = []
        for k, v in code.input_policy_init_kwargs.items():
            # We would need the verify key of the data scientist to be able to index the correct client
            # Since we do not want the data scientist to pass in the clients to the enclave client
            # from a UX perspecitve.
            # we will use the recent node id to find the correct client
            # assuming that it is the correct client
            # Warning: This could lead to inconsistent results, when we have multiple clients
            # in the same node pointing to the same node.
            # One way, by which we could solve this in the long term,
            # by forcing the user to pass only assets to the sy.ExactMatch,
            # by which we could extract the verify key of the data scientist
            # as each object comes with a verify key and node_uid
            # the asset object would contain the verify key of the data scientist.
            api = APIRegistry.get_by_recent_node_uid(k.node_id)
            if api is None:
                raise ValueError(f"could not find client for input {v}")
            else:
                apis += [api]

        for api in apis:
            res = api.services.code.request_code_execution(code=code)
            if isinstance(res, SyftError):
                return res

        # we are using the real method here, see the .code property getter
        _ = self.code
        res = self._request_code_execution(code=code)

        return res

    def _repr_html_(self) -> str:
        commands = """
        <li><span class='syft-code-block'>&lt;your_client&gt;
        .request_code_execution</span> - submit code to enclave for execution</li>
        """

        command_list = f"""
        <ul style='padding-left: 1em;'>
            {commands}
        </ul>
        """

        small_grid_symbol_logo = base64read("small-grid-symbol-logo.png")

        url = getattr(self.connection, "url", None)
        node_details = f"<strong>URL:</strong> {url}<br />" if url else ""
        node_details += (
            f"<strong>Node Type:</strong> {self.metadata.node_type.capitalize()}<br />"
        )
        node_side_type = (
            "Low Side"
            if self.metadata.node_side_type == NodeSideType.LOW_SIDE.value
            else "High Side"
        )
        node_details += f"<strong>Node Side Type:</strong> {node_side_type}<br />"
        node_details += (
            f"<strong>Syft Version:</strong> {self.metadata.syft_version}<br />"
        )

        return f"""
        <style>
            {fonts_css}

            .syft-container {{
                padding: 5px;
                font-family: 'Open Sans';
            }}
            .syft-alert-info {{
                color: #1F567A;
                background-color: #C2DEF0;
                border-radius: 4px;
                padding: 5px;
                padding: 13px 10px
            }}
            .syft-code-block {{
                background-color: #f7f7f7;
                border: 1px solid #cfcfcf;
                padding: 0px 2px;
            }}
            .syft-space {{
                margin-top: 1em;
            }}
        </style>
        <div class="syft-client syft-container">
            <img src="{small_grid_symbol_logo}" alt="Logo"
            style="width:48px;height:48px;padding:3px;">
            <h2>Welcome to {self.name}</h2>
            <div class="syft-space">
                {node_details}
            </div>
            <div class='syft-alert-info syft-space'>
                &#9432;&nbsp;
                This node is run by the library PySyft to learn more about how it works visit
                <a href="https://github.com/OpenMined/PySyft">github.com/OpenMined/PySyft</a>.
            </div>
            <h4>Commands to Get Started</h4>
            {command_list}
        </div><br />
        """



# stdlib
from enum import Enum
import hashlib
import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...abstract_node import NodeSideType
from ...client.api import APIRegistry
from ...custom_worker.config import WorkerConfig
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...serde.serialize import _serialize
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.syft_migration import migrate
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import add_node_uid_for_key
from ...types.transforms import drop
from ...types.transforms import generate_id
from ...types.transforms import make_set_default
from ...types.transforms import transform
from ...types.twin_object import TwinObject
from ...types.uid import LineageID
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.markdown import markdown_as_class_with_fields
from ...util.notebook_ui.notebook_addons import REQUEST_ICON
from ...util.util import prompt_warning_message
from ..action.action_object import ActionObject
from ..action.action_service import ActionService
from ..action.action_store import ActionObjectPermission
from ..action.action_store import ActionPermission
from ..blob_storage.service import BlobStorageService
from ..code.user_code import UserCode
from ..code.user_code import UserCodeStatus
from ..context import AuthedServiceContext
from ..context import ChangeContext
from ..job.job_stash import Job
from ..job.job_stash import JobInfo
from ..job.job_stash import JobStatus
from ..notification.notifications import Notification
from ..response import SyftError
from ..response import SyftSuccess
from ..user.user import UserView


@serializable()
class RequestStatus(Enum):
    PENDING = 0
    REJECTED = 1
    APPROVED = 2


@serializable()
class Change(SyftObject):
    __canonical_name__ = "Change"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_obj: Optional[LinkedObject]

    def is_type(self, type_: type) -> bool:
        return self.linked_obj and type_ == self.linked_obj.object_type


@serializable()
class ChangeStatus(SyftObject):
    __canonical_name__ = "ChangeStatus"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    change_id: UID
    applied: bool = False

    @classmethod
    def from_change(cls, change: Change, applied: bool) -> Self:
        return cls(change_id=change.id, applied=applied)


@serializable()
class ActionStoreChange(Change):
    __canonical_name__ = "ActionStoreChange"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_obj: LinkedObject
    apply_permission_type: ActionPermission

    __repr_attrs__ = ["linked_obj", "apply_permission_type"]

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            action_service: ActionService = context.node.get_service(ActionService)
            blob_storage_service = context.node.get_service(BlobStorageService)
            action_store = action_service.store

            # can we ever have a lineage ID in the store?
            obj_uid = self.linked_obj.object_uid
            obj_uid = obj_uid.id if isinstance(obj_uid, LineageID) else obj_uid

            action_obj = action_store.get(
                uid=obj_uid,
                credentials=context.approving_user_credentials,
            )

            if action_obj.is_err():
                return Err(SyftError(message=f"{action_obj.err()}"))

            action_obj = action_obj.ok()

            owner_permission = ActionObjectPermission(
                uid=obj_uid,
                credentials=context.approving_user_credentials,
                permission=self.apply_permission_type,
            )
            if action_store.has_permission(permission=owner_permission):
                id_action = (
                    action_obj.id
                    if not isinstance(action_obj.id, LineageID)
                    else action_obj.id.id
                )
                requesting_permission_action_obj = ActionObjectPermission(
                    uid=id_action,
                    credentials=context.requesting_user_credentials,
                    permission=self.apply_permission_type,
                )
                if isinstance(action_obj, TwinObject):
                    uid_blob = action_obj.private.syft_blob_storage_entry_id
                else:
                    uid_blob = action_obj.syft_blob_storage_entry_id
                requesting_permission_blob_obj = ActionObjectPermission(
                    uid=uid_blob,
                    credentials=context.requesting_user_credentials,
                    permission=self.apply_permission_type,
                )
                if apply:
                    action_store.add_permission(requesting_permission_action_obj)
                    blob_storage_service.stash.add_permission(
                        requesting_permission_blob_obj
                    )
                else:
                    if action_store.has_permission(requesting_permission_action_obj):
                        action_store.remove_permission(requesting_permission_action_obj)
                    if blob_storage_service.stash.has_permission(
                        requesting_permission_blob_obj
                    ):
                        blob_storage_service.stash.remove_permission(
                            requesting_permission_blob_obj
                        )
            else:
                return Err(
                    SyftError(
                        message=f"No permission for approving_user_credentials {context.approving_user_credentials}"
                    )
                )
            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            print(f"failed to apply {type(self)}", e)
            return Err(SyftError(message=str(e)))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self):
        return f"Apply <b>{self.apply_permission_type}</b> to \
            <i>{self.linked_obj.object_type.__canonical_name__}:{self.linked_obj.object_uid.short()}</i>"


@serializable()
class CreateCustomImageChange(Change):
    __canonical_name__ = "CreateCustomImageChange"
    __version__ = SYFT_OBJECT_VERSION_1

    config: WorkerConfig
    tag: str

    __repr_attrs__ = ["config", "tag"]

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            worker_image_service = context.node.get_service("SyftWorkerImageService")

            service_context = context.to_service_ctx()
            result = worker_image_service.submit_dockerfile(
                service_context, docker_config=self.config
            )

            if isinstance(result, SyftError):
                return Err(result)

            result = worker_image_service.stash.get_by_docker_config(
                service_context.credentials, config=self.config
            )

            if result.is_err():
                return Err(SyftError(message=f"{result.err()}"))

            worker_image = result.ok()

            build_result = worker_image_service.build(
                service_context, image_uid=worker_image.id, tag=self.tag
            )
            return Ok(build_result)

        except Exception as e:
            return Err(SyftError(message=f"Failed to create/build image: {e}"))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self):
        return f"Create Image for Config: {self.config} with tag: {self.tag}"


@serializable()
class CreateCustomWorkerPoolChange(Change):
    __canonical_name__ = "CreateCustomWorkerPoolChange"
    __version__ = SYFT_OBJECT_VERSION_1

    pool_name: str
    num_workers: int
    image_uid: Optional[UID]
    config: Optional[WorkerConfig]

    __repr_attrs__ = ["pool_name", "num_workers", "image_uid"]

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        """
        This function is run when the DO approves (apply=True)
        or deny (apply=False) the request.
        """
        if apply:
            # get the worker pool service and try to launch a pool
            worker_pool_service = context.node.get_service("SyftWorkerPoolService")
            service_context: AuthedServiceContext = context.to_service_ctx()

            if self.config is not None:
                result = worker_pool_service.image_stash.get_by_docker_config(
                    service_context.credentials, self.config
                )
                if result.is_err():
                    return Err(SyftError(message=f"{result.err()}"))
                worker_image = result.ok()
                self.image_uid = worker_image.id

            result = worker_pool_service.launch(
                context=service_context,
                name=self.pool_name,
                image_uid=self.image_uid,
                num_workers=self.num_workers,
            )
            if isinstance(result, SyftError):
                return Err(result)
            else:
                return Ok(result)
        else:
            return Err(
                SyftError(
                    message=f"Request to create a worker pool with name {self.name} denied"
                )
            )

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self):
        return (
            f"Create Worker Pool '{self.pool_name}' for Image with id {self.image_uid}"
        )


@serializable()
class Request(SyftObject):
    __canonical_name__ = "Request"
    __version__ = SYFT_OBJECT_VERSION_1

    requesting_user_verify_key: SyftVerifyKey
    requesting_user_name: str = ""
    requesting_user_email: Optional[str] = ""
    requesting_user_institution: Optional[str] = ""
    approving_user_verify_key: Optional[SyftVerifyKey]
    request_time: DateTime
    updated_at: Optional[DateTime]
    node_uid: UID
    request_hash: str
    changes: List[Change]
    history: List[ChangeStatus] = []

    __attr_searchable__ = [
        "requesting_user_verify_key",
        "approving_user_verify_key",
    ]
    __attr_unique__ = ["request_hash"]
    __repr_attrs__ = [
        "request_time",
        "updated_at",
        "status",
        "changes",
        "requesting_user_verify_key",
    ]

    def _repr_html_(self) -> Any:
        # add changes
        updated_at_line = ""
        if self.updated_at is not None:
            updated_at_line += (
                f"<p><strong>Created by: </strong>{self.requesting_user_name}</p>"
            )
        str_changes = []
        for change in self.changes:
            str_change = (
                change.__repr_syft_nested__()
                if hasattr(change, "__repr_syft_nested__")
                else type(change)
            )
            str_change = f"{str_change}. "
            str_changes.append(str_change)
        str_changes = "\n".join(str_changes)
        api = APIRegistry.api_for(
            self.node_uid,
            self.syft_client_verify_key,
        )
        shared_with_line = ""
        if self.code and len(self.code.output_readers) > 0:
            # owner_names = ["canada", "US"]
            owners_string = " and ".join(
                [f"<strong>{x}</strong>" for x in self.code.output_reader_names]
            )
            shared_with_line += (
                f"<p><strong>Custom Policy: </strong> "
                f"outputs are <strong>shared</strong> with the owners of {owners_string} once computed"
            )

        metadata = api.services.metadata.get_metadata()
        node_name = api.node_name.capitalize() if api.node_name is not None else ""

        email_str = (
            f"({self.requesting_user_email})" if self.requesting_user_email else ""
        )
        institution_str = (
            f"<strong>Institution:</strong> {self.requesting_user_institution}"
            if self.requesting_user_institution
            else ""
        )

        return f"""
            <style>
            .syft-request {{color: {SURFACE[options.color_theme]};}}
            </style>
            <div class='syft-request'>
                <h3>Request</h3>
                <p><strong>Id: </strong>{self.id}</p>
                <p><strong>Request time: </strong>{self.request_time}</p>
                {updated_at_line}
                {shared_with_line}
                <p><strong>Status: </strong>{self.status}</p>
                <p><strong>Requested on: </strong> {node_name} of type <strong> \
                    {metadata.node_type.value.capitalize()}</strong></p>
                <p><strong>Requested by:</strong> {self.requesting_user_name} {email_str} {institution_str}</p>
                <p><strong>Changes: </strong> {str_changes}</p>
            </div>

            """

    def _coll_repr_(self):
        if self.status == RequestStatus.APPROVED:
            badge_color = "badge-green"
        elif self.status == RequestStatus.PENDING:
            badge_color = "badge-gray"
        else:
            badge_color = "badge-red"

        status_badge = {"value": self.status.name.capitalize(), "type": badge_color}

        user_data = [
            self.requesting_user_name,
            self.requesting_user_email,
            self.requesting_user_institution,
        ]

        return {
            "Description": " ".join([x.__repr_syft_nested__() for x in self.changes]),
            "Requested By": "\n".join(user_data),
            "Status": status_badge,
        }

    @property
    def codes(self) -> Any:
        for change in self.changes:
            if isinstance(change, UserCodeStatusChange):
                return change.codes
        return SyftError(
            message="This type of request does not have code associated with it."
        )

    @property
    def code(self) -> Any:
        for change in self.changes:
            if isinstance(change, UserCodeStatusChange):
                return change.code
        return SyftError(
            message="This type of request does not have code associated with it."
        )

    def get_results(self) -> Any:
        return self.code.get_results()

    @property
    def current_change_state(self) -> Dict[UID, bool]:
        change_applied_map = {}
        for change_status in self.history:
            # only store the last change
            change_applied_map[change_status.change_id] = change_status.applied

        return change_applied_map

    @property
    def icon(self):
        return REQUEST_ICON

    @property
    def status(self) -> RequestStatus:
        if len(self.history) == 0:
            return RequestStatus.PENDING

        all_changes_applied = all(self.current_change_state.values()) and (
            len(self.current_change_state) == len(self.changes)
        )

        request_status = (
            RequestStatus.APPROVED if all_changes_applied else RequestStatus.REJECTED
        )

        return request_status

    def approve(self, disable_warnings: bool = False, approve_nested: bool = False):
        api = APIRegistry.api_for(
            self.node_uid,
            self.syft_client_verify_key,
        )
        # TODO: Refactor so that object can also be passed to generate warnings
        metadata = api.connection.get_node_metadata(api.signing_key)
        message, is_enclave = None, False

        is_code_request = not isinstance(self.codes, SyftError)

        if is_code_request and len(self.codes) > 1 and not approve_nested:
            return SyftError(
                message="Multiple codes detected, please use approve_nested=True"
            )

        if self.code and not isinstance(self.code, SyftError):
            is_enclave = getattr(self.code, "enclave_metadata", None) is not None

        if is_enclave:
            message = "On approval, the result will be released to the enclave."
        elif metadata.node_side_type == NodeSideType.HIGH_SIDE.value:
            message = (
                "You're approving a request on "
                f"{metadata.node_side_type} side {metadata.node_type} "
                "which may host datasets with private information."
            )
        if message and metadata.show_warnings and not disable_warnings:
            prompt_warning_message(message=message, confirm=True)

        print(f"Approving request for domain {api.node_name}")
        return api.services.request.apply(self.id)

    def deny(self, reason: str):
        """Denies the particular request.

        Args:
            reason (str): Reason for which the request has been denied.
        """
        api = APIRegistry.api_for(
            self.node_uid,
            self.syft_client_verify_key,
        )
        return api.services.request.undo(uid=self.id, reason=reason)

    def approve_with_client(self, client):
        print(f"Approving request for domain {client.name}")
        return client.api.services.request.apply(self.id)

    def apply(self, context: AuthedServiceContext) -> Result[SyftSuccess, SyftError]:
        change_context = ChangeContext.from_service(context)
        change_context.requesting_user_credentials = self.requesting_user_verify_key
        for change in self.changes:
            # by default change status is not applied
            change_status = ChangeStatus(change_id=change.id, applied=False)
            result = change.apply(context=change_context)
            if isinstance(result, SyftError):
                return result
            if result.is_err():
                # add to history and save history to request
                self.history.append(change_status)
                self.save(context=context)
                return result

            # If no error, then change successfully applied.
            change_status.applied = True
            self.history.append(change_status)

        self.updated_at = DateTime.now()
        self.save(context=context)

        return Ok(SyftSuccess(message=f"Request {self.id} changes applied"))

    def undo(self, context: AuthedServiceContext) -> Result[SyftSuccess, SyftError]:
        change_context = ChangeContext.from_service(context)
        change_context.requesting_user_credentials = self.requesting_user_verify_key

        current_change_state = self.current_change_state
        for change in self.changes:
            # by default change status is not applied
            is_change_applied = current_change_state.get(change.id, False)
            change_status = ChangeStatus(
                change_id=change.id,
                applied=is_change_applied,
            )
            # undo here may be deny for certain Changes (UserCodeChange)
            result = change.undo(context=change_context)
            if result.is_err():
                # add to history and save history to request
                self.history.append(change_status)
                self.save(context=context)
                return result

            # If no error, then change successfully undone.
            change_status.applied = False
            self.history.append(change_status)

        self.updated_at = DateTime.now()
        result = self.save(context=context)
        if isinstance(result, SyftError):
            return Err(result)
        # override object with latest changes.
        self = result
        return Ok(SyftSuccess(message=f"Request {self.id} changes undone."))

    def save(self, context: AuthedServiceContext) -> Result[SyftSuccess, SyftError]:
        # relative
        from .request_service import RequestService

        save_method = context.node.get_service_method(RequestService.save)
        return save_method(context=context, request=self)

    def _get_latest_or_create_job(self) -> Union[Job, SyftError]:
        """Get the latest job for this requests user_code, or creates one if no jobs exist"""
        api = APIRegistry.api_for(self.node_uid, self.syft_client_verify_key)
        job_service = api.services.job

        existing_jobs = job_service.get_by_user_code_id(self.code.id)
        if isinstance(existing_jobs, SyftError):
            return existing_jobs

        if len(existing_jobs) == 0:
            job = job_service.create_job_for_user_code_id(self.code.id)
        else:
            job = existing_jobs[-1]

        return job

    def accept_by_depositing_result(self, result: Any, force: bool = False):
        # this code is extremely brittle because its a work around that relies on
        # the type of request being very specifically tied to code which needs approving

        # Special case for results from Jobs (High-low side async)
        if isinstance(result, JobInfo):
            job_info = result
            if not job_info.includes_result:
                return SyftError(
                    message="JobInfo should not include result. Use sync_job instead."
                )
            result = job_info.result
        else:
            # NOTE result is added at the end of function (once ActionObject is created)
            job_info = JobInfo(
                includes_metadata=True,
                includes_result=True,
                status=JobStatus.COMPLETED,
                resolved=True,
            )

        change = self.changes[0]
        if not change.is_type(UserCode):
            raise TypeError(
                f"accept_by_depositing_result can only be run on {UserCode} not "
                f"{change.linked_obj.object_type}"
            )
        if not type(change) == UserCodeStatusChange:
            raise TypeError(
                f"accept_by_depositing_result can only be run on {UserCodeStatusChange} not "
                f"{type(change)}"
            )

        api = APIRegistry.api_for(self.node_uid, self.syft_client_verify_key)
        if not api:
            raise Exception(f"Login to {self.node_uid} first.")

        is_approved = change.approved

        permission_request = self.approve()
        if isinstance(permission_request, SyftError):
            return permission_request

        code = change.linked_obj.resolve
        state = code.output_policy

        # This weird order is due to the fact that state is None before calling approve
        # we could fix it in a future release
        if is_approved:
            if not force:
                return SyftError(
                    message="Already approved, if you want to force updating the result use force=True"
                )
            action_obj_id = state.output_history[0].outputs[0]
            action_object = ActionObject.from_obj(
                result,
                id=action_obj_id,
                syft_client_verify_key=api.signing_key.verify_key,
                syft_node_location=api.node_uid,
            )
            blob_store_result = action_object._save_to_blob_storage()
            if isinstance(blob_store_result, SyftError):
                return blob_store_result
            result = api.services.action.set(action_object)
            if isinstance(result, SyftError):
                return result
        else:
            action_object = ActionObject.from_obj(
                result,
                syft_client_verify_key=api.signing_key.verify_key,
                syft_node_location=api.node_uid,
            )
            blob_store_result = action_object._save_to_blob_storage()
            if isinstance(blob_store_result, SyftError):
                return blob_store_result
            result = api.services.action.set(action_object)
            if isinstance(result, SyftError):
                return result

            ctx = AuthedServiceContext(credentials=api.signing_key.verify_key)

            state.apply_output(context=ctx, outputs=result)
            policy_state_mutation = ObjectMutation(
                linked_obj=change.linked_obj,
                attr_name="output_policy",
                match_type=True,
                value=state,
            )

            action_object_link = LinkedObject.from_obj(result, node_uid=self.node_uid)
            permission_change = ActionStoreChange(
                linked_obj=action_object_link,
                apply_permission_type=ActionPermission.READ,
            )

            new_changes = [policy_state_mutation, permission_change]
            result_request = api.services.request.add_changes(
                uid=self.id, changes=new_changes
            )
            if isinstance(result_request, SyftError):
                return result_request
            self = result_request

            approved = self.approve(disable_warnings=True)
            if isinstance(approved, SyftError):
                return approved

        job_info.result = action_object
        job = self._get_latest_or_create_job()
        job.apply_info(job_info)

        api = APIRegistry.api_for(self.node_uid, self.syft_client_verify_key)
        job_service = api.services.job
        res = job_service.update(job)
        if isinstance(res, SyftError):
            return res

        return SyftSuccess(message="Request submitted for updating result.")

    def sync_job(self, job_info: JobInfo, **kwargs) -> Result[SyftSuccess, SyftError]:
        if job_info.includes_result:
            return SyftError(
                message="This JobInfo includes a Result. Please use Request.accept_by_depositing_result instead."
            )

        api = APIRegistry.api_for(self.node_uid, self.syft_client_verify_key)
        job_service = api.services.job

        job = self._get_latest_or_create_job()
        job.apply_info(job_info)
        return job_service.update(job)


@serializable()
class RequestInfo(SyftObject):
    # version
    __canonical_name__ = "RequestInfo"
    __version__ = SYFT_OBJECT_VERSION_1

    user: UserView
    request: Request
    message: Notification


@serializable()
class RequestInfoFilter(SyftObject):
    # version
    __canonical_name__ = "RequestInfoFilter"
    __version__ = SYFT_OBJECT_VERSION_1

    name: Optional[str]


@serializable()
class SubmitRequest(SyftObject):
    __canonical_name__ = "SubmitRequest"
    __version__ = SYFT_OBJECT_VERSION_1

    changes: List[Change]
    requesting_user_verify_key: Optional[SyftVerifyKey]


def hash_changes(context: TransformContext) -> TransformContext:
    request_time = context.output["request_time"]
    key = context.output["requesting_user_verify_key"]
    changes = context.output["changes"]

    time_hash = hashlib.sha256(
        _serialize(request_time.utc_timestamp, to_bytes=True)
    ).digest()
    key_hash = hashlib.sha256(bytes(key.verify_key)).digest()
    changes_hash = hashlib.sha256(_serialize(changes, to_bytes=True)).digest()
    final_hash = hashlib.sha256(time_hash + key_hash + changes_hash).hexdigest()

    context.output["request_hash"] = final_hash
    return context


def add_request_time(context: TransformContext) -> TransformContext:
    context.output["request_time"] = DateTime.now()
    return context


def check_requesting_user_verify_key(context: TransformContext) -> TransformContext:
    if context.obj.requesting_user_verify_key and context.node.is_root(
        context.credentials
    ):
        context.output[
            "requesting_user_verify_key"
        ] = context.obj.requesting_user_verify_key
    else:
        context.output["requesting_user_verify_key"] = context.credentials
    return context


def add_requesting_user_info(context: TransformContext) -> TransformContext:
    try:
        user_key = context.output["requesting_user_verify_key"]
        user_service = context.node.get_service("UserService")
        user = user_service.get_by_verify_key(user_key)
        context.output["requesting_user_name"] = user.name
        context.output["requesting_user_email"] = user.email
        context.output["requesting_user_institution"] = (
            user.institution if user.institution else ""
        )
    except Exception:
        context.output["requesting_user_name"] = "guest_user"
    return context


@transform(SubmitRequest, Request)
def submit_request_to_request() -> List[Callable]:
    return [
        generate_id,
        add_node_uid_for_key("node_uid"),
        add_request_time,
        check_requesting_user_verify_key,
        add_requesting_user_info,
        hash_changes,
    ]


@serializable()
class ObjectMutation(Change):
    __canonical_name__ = "ObjectMutation"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_obj: Optional[LinkedObject]
    attr_name: str
    value: Optional[Any]
    match_type: bool
    previous_value: Optional[Any]

    __repr_attrs__ = ["linked_obj", "attr_name"]

    def mutate(self, obj: Any, value: Optional[Any]) -> Any:
        # check if attribute is a property setter first
        # this seems necessary for pydantic types
        attr = getattr(type(obj), self.attr_name, None)
        if inspect.isdatadescriptor(attr):
            self.previous_value = attr.fget(obj)
            attr.fset(obj, value)

        else:
            self.previous_value = getattr(obj, self.attr_name, None)
            setattr(obj, self.attr_name, value)
        return obj

    def __repr_syft_nested__(self):
        return f"Mutate <b>{self.attr_name}</b> to <b>{self.value}</b>"

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            obj = self.linked_obj.resolve_with_context(context)
            if obj.is_err():
                return Err(SyftError(message=obj.err()))
            obj = obj.ok()
            if apply:
                obj = self.mutate(obj, value=self.value)
                self.linked_obj.update_with_context(context, obj)
            else:
                # unset the set value
                obj = self.mutate(obj, value=self.previous_value)
                self.linked_obj.update_with_context(context, obj)

            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            print(f"failed to apply {type(self)}. {e}")
            return Err(SyftError(message=str(e)))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)


def type_for_field(object_type: type, attr_name: str) -> Optional[type]:
    field_type = None
    try:
        field_type = object_type.__dict__["__annotations__"][attr_name]
    except Exception:  # nosec
        try:
            field_type = object_type.__fields__.get(attr_name, None).type_
        except Exception:  # nosec
            pass
    return field_type


@serializable()
class EnumMutation(ObjectMutation):
    __canonical_name__ = "EnumMutation"
    __version__ = SYFT_OBJECT_VERSION_1

    enum_type: Type[Enum]
    value: Optional[Enum]
    match_type: bool = True

    __repr_attrs__ = ["linked_obj", "attr_name", "value"]

    @property
    def valid(self) -> Union[SyftSuccess, SyftError]:
        if self.match_type and not isinstance(self.value, self.enum_type):
            return SyftError(
                message=f"{type(self.value)} must be of type: {self.enum_type}"
            )
        return SyftSuccess(message=f"{type(self)} valid")

    @staticmethod
    def from_obj(
        linked_obj: LinkedObject, attr_name: str, value: Optional[Enum] = None
    ) -> Self:
        enum_type = type_for_field(linked_obj.object_type, attr_name)
        return EnumMutation(
            linked_obj=linked_obj,
            attr_name=attr_name,
            enum_type=enum_type,
            value=value,
            match_type=True,
        )

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            valid = self.valid
            if not valid:
                return Err(valid)
            obj = self.linked_obj.resolve_with_context(context)
            if obj.is_err():
                return SyftError(message=obj.err())
            obj = obj.ok()
            if apply:
                obj = self.mutate(obj)

                self.linked_obj.update_with_context(context, obj)
            else:
                raise NotImplementedError
            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            print(f"failed to apply {type(self)}. {e}")
            return Err(SyftError(message=e))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    def __repr_syft_nested__(self):
        return f"Mutate <b>{self.enum_type}</b> to <b>{self.value}</b>"

    @property
    def link(self) -> Optional[SyftObject]:
        if self.linked_obj:
            return self.linked_obj.resolve
        return None


@serializable()
class UserCodeStatusChangeV1(Change):
    __canonical_name__ = "UserCodeStatusChange"
    __version__ = SYFT_OBJECT_VERSION_1

    value: UserCodeStatus
    linked_obj: LinkedObject
    match_type: bool = True
    __repr_attrs__ = [
        "link.service_func_name",
        "link.input_policy_type.__canonical_name__",
        "link.output_policy_type.__canonical_name__",
        "link.status.approved",
    ]


@serializable()
class UserCodeStatusChange(Change):
    __canonical_name__ = "UserCodeStatusChange"
    __version__ = SYFT_OBJECT_VERSION_2

    value: UserCodeStatus
    linked_obj: LinkedObject
    nested_solved: bool = False
    match_type: bool = True
    __repr_attrs__ = [
        "link.service_func_name",
        "link.input_policy_type.__canonical_name__",
        "link.output_policy_type.__canonical_name__",
        "link.worker_pool_id",
        "link.status.approved",
    ]

    @property
    def code(self):
        return self.link

    @property
    def codes(self):
        def recursive_code(node):
            codes = []
            for _, (obj, new_node) in node.items():
                codes.append(obj.resolve)
                codes.extend(recursive_code(new_node))
            return codes

        codes = [self.link]
        codes.extend(recursive_code(self.link.nested_codes))
        return codes

    def nested_repr(self, node=None, level=0):
        msg = ""
        if node is None:
            node = self.link.nested_codes
        for service_func_name, (_, new_node) in node.items():
            msg = "├──" + "──" * level + f"{service_func_name}<br>"
            msg += self.nested_repr(node=new_node, level=level + 1)
        return msg

    def __repr_syft_nested__(self):
        msg = f"Request to change <b>{self.link.service_func_name}</b> (Pool Id: <b>{self.link.worker_pool_id}</b>) "
        msg += "to permission <b>RequestStatus.APPROVED</b>"
        if self.nested_solved:
            if self.link.nested_codes == {}:
                msg += ". No nested requests"
            else:
                msg += ".<br><br>This change requests the following nested functions calls:<br>"
                msg += self.nested_repr()
        else:
            msg += ". Nested Requests not resolved"
        return msg

    def _repr_markdown_(self) -> str:
        link = self.link
        input_policy_type = (
            link.input_policy_type.__canonical_name__
            if link.input_policy_type is not None
            else None
        )
        output_policy_type = (
            link.output_policy_type.__canonical_name__
            if link.output_policy_type is not None
            else None
        )
        repr_dict = {
            "function": link.service_func_name,
            "input_policy_type": f"{input_policy_type}",
            "output_policy_type": f"{output_policy_type}",
            "approved": f"{link.status.approved}",
        }
        return markdown_as_class_with_fields(self, repr_dict)

    @property
    def approved(self) -> bool:
        return self.linked_obj.resolve.status.approved

    @property
    def valid(self) -> Union[SyftSuccess, SyftError]:
        if self.match_type and not isinstance(self.value, UserCodeStatus):
            return SyftError(
                message=f"{type(self.value)} must be of type: {UserCodeStatus}"
            )
        return SyftSuccess(message=f"{type(self)} valid")

    # def get_nested_requests(self, context, code_tree: Dict[str: Tuple[LinkedObject, Dict]]):
    #     approved_nested_codes = {}
    #     for key, (linked_obj, new_code_tree) in code_tree.items():
    #         code_obj = linked_obj.resolve_with_context(context).ok()
    #         approved_nested_codes[key] = code_obj.id

    #         res = self.get_nested_requests(context, new_code_tree)
    #         if isinstance(res, SyftError):
    #             return res
    #         code_obj.nested_codes = res
    #         linked_obj.update_with_context(context, code_obj)

    #     return approved_nested_codes

    def mutate(self, obj: UserCode, context: ChangeContext, undo: bool) -> Any:
        reason: str = context.extra_kwargs.get("reason", "")
        if not undo:
            res = obj.status.mutate(
                value=(self.value, reason),
                node_name=context.node.name,
                node_id=context.node.id,
                verify_key=context.node.signing_key.verify_key,
            )
            if isinstance(res, SyftError):
                return res
        else:
            res = obj.status.mutate(
                value=(UserCodeStatus.DENIED, reason),
                node_name=context.node.name,
                node_id=context.node.id,
                verify_key=context.node.signing_key.verify_key,
            )
        if not isinstance(res, SyftError):
            obj.status = res
            return obj
        return res

    def is_enclave_request(self, user_code: UserCode):
        return (
            user_code.is_enclave_code is not None
            and self.value == UserCodeStatus.APPROVED
        )

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[SyftSuccess, SyftError]:
        try:
            valid = self.valid
            if not valid:
                return Err(valid)
            obj = self.linked_obj.resolve_with_context(context)
            if obj.is_err():
                return Err(SyftError(message=obj.err()))
            obj = obj.ok()
            if apply:
                res = self.mutate(obj, context, undo=False)

                if isinstance(res, SyftError):
                    return Err(res.message)

                # relative
                from ..enclave.enclave_service import propagate_inputs_to_enclave

                user_code = res

                if self.is_enclave_request(user_code):
                    enclave_res = propagate_inputs_to_enclave(
                        user_code=res, context=context
                    )
                    if isinstance(enclave_res, SyftError):
                        return enclave_res
                self.linked_obj.update_with_context(context, user_code)
            else:
                res = self.mutate(obj, context, undo=True)
                if isinstance(res, SyftError):
                    return Err(res.message)

                # TODO: Handle Enclave approval.
                self.linked_obj.update_with_context(context, res)
            return Ok(SyftSuccess(message=f"{type(self)} Success"))
        except Exception as e:
            print(f"failed to apply {type(self)}. {e}")
            return Err(SyftError(message=str(e)))

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context=context, apply=False)

    @property
    def link(self) -> Optional[SyftObject]:
        if self.linked_obj:
            return self.linked_obj.resolve
        return None


@migrate(UserCodeStatusChange, UserCodeStatusChangeV1)
def downgrade_usercodestatuschange_v2_to_v1():
    return [
        drop("nested_solved"),
    ]


@migrate(UserCodeStatusChangeV1, UserCodeStatusChange)
def upgrade_usercodestatuschange_v1_to_v2():
    return [
        make_set_default("nested_solved", True),
    ]



# future
from __future__ import annotations

# stdlib
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
import sqlite3
import tempfile
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from pydantic import Field
from pydantic import validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ..serde.deserialize import _deserialize
from ..serde.serializable import serializable
from ..serde.serialize import _serialize
from ..types.uid import UID
from ..util.util import thread_ident
from .document_store import DocumentStore
from .document_store import PartitionSettings
from .document_store import StoreClientConfig
from .document_store import StoreConfig
from .kv_document_store import KeyValueBackingStore
from .kv_document_store import KeyValueStorePartition
from .locks import FileLockingConfig
from .locks import LockingConfig
from .locks import SyftLock

# here we can create a single connection per cache_key
# since pytest is concurrent processes, we need to isolate each connection
# by its filename and optionally the thread that its running in
# we keep track of each SQLiteBackingStore init in REF_COUNTS
# when it hits 0 we can close the connection and release the file descriptor
SQLITE_CONNECTION_POOL_DB: Dict[str, sqlite3.Connection] = {}
SQLITE_CONNECTION_POOL_CUR: Dict[str, sqlite3.Cursor] = {}
REF_COUNTS: Dict[str, int] = defaultdict(int)


def cache_key(db_name: str) -> str:
    return f"{db_name}_{thread_ident()}"


def _repr_debug_(value: Any) -> str:
    if hasattr(value, "_repr_debug_"):
        return str(value._repr_debug_())
    return repr(value)


def raise_exception(table_name: str, e: Exception):
    if "disk I/O error" in str(e):
        message = f"Error usually related to concurrent writes. {str(e)}"
        raise Exception(message)

    if "Cannot operate on a closed database" in str(e):
        message = (
            "Error usually related to calling self.db.close()"
            + f"before last SQLiteBackingStore.__del__ gets called. {str(e)}"
        )
        raise Exception(message)

    # if its something else other than "table already exists" raise original e
    if f"table {table_name} already exists" not in str(e):
        raise e


@serializable(attrs=["index_name", "settings", "store_config"])
class SQLiteBackingStore(KeyValueBackingStore):
    """Core Store logic for the SQLite stores.

    Parameters:
        `index_name`: str
            Index name
        `settings`: PartitionSettings
            Syft specific settings
        `store_config`: SQLiteStoreConfig
            Connection Configuration
        `ddtype`: Type
            Class used as fallback on `get` errors
    """

    def __init__(
        self,
        index_name: str,
        settings: PartitionSettings,
        store_config: StoreConfig,
        ddtype: Optional[type] = None,
    ) -> None:
        self.index_name = index_name
        self.settings = settings
        self.store_config = store_config
        self._ddtype = ddtype
        self.file_path = self.store_config.client_config.file_path
        self.db_filename = store_config.client_config.filename

        # if tempfile.TemporaryDirectory() varies from process to process
        # could this cause different locks on the same file
        temp_dir = tempfile.TemporaryDirectory().name
        lock_path = Path(temp_dir) / "sqlite_locks" / self.db_filename
        self.lock_config = FileLockingConfig(client_path=lock_path)
        self.create_table()
        REF_COUNTS[cache_key(self.db_filename)] += 1

    @property
    def table_name(self) -> str:
        return f"{self.settings.name}_{self.index_name}"

    def _connect(self) -> None:
        # SQLite is not thread safe by default so we ensure that each connection
        # comes from a different thread. In cases of Uvicorn and other AWSGI servers
        # there will be many threads handling incoming requests so we need to ensure
        # that different connections are used in each thread. By using a dict for the
        # _db and _cur we can ensure they are never shared

        path = Path(self.file_path)
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        connection = sqlite3.connect(
            self.file_path,
            timeout=self.store_config.client_config.timeout,
            check_same_thread=False,  # do we need this if we use the lock?
            # check_same_thread=self.store_config.client_config.check_same_thread,
        )
        # TODO: Review OSX compatibility.
        # Set journal mode to WAL.
        # connection.execute("pragma journal_mode=wal")
        SQLITE_CONNECTION_POOL_DB[cache_key(self.db_filename)] = connection

    def create_table(self) -> None:
        try:
            with SyftLock(self.lock_config):
                self.cur.execute(
                    f"create table {self.table_name} (uid VARCHAR(32) NOT NULL PRIMARY KEY, "  # nosec
                    + "repr TEXT NOT NULL, value BLOB NOT NULL, "  # nosec
                    + "sqltime TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL)"  # nosec
                )
                self.db.commit()
        except Exception as e:
            raise_exception(self.table_name, e)

    @property
    def db(self) -> sqlite3.Connection:
        if cache_key(self.db_filename) not in SQLITE_CONNECTION_POOL_DB:
            self._connect()
        return SQLITE_CONNECTION_POOL_DB[cache_key(self.db_filename)]

    @property
    def cur(self) -> sqlite3.Cursor:
        if cache_key(self.db_filename) not in SQLITE_CONNECTION_POOL_CUR:
            SQLITE_CONNECTION_POOL_CUR[cache_key(self.db_filename)] = self.db.cursor()

        return SQLITE_CONNECTION_POOL_CUR[cache_key(self.db_filename)]

    def _close(self) -> None:
        self._commit()
        REF_COUNTS[cache_key(self.db_filename)] -= 1
        if REF_COUNTS[cache_key(self.db_filename)] <= 0:
            # once you close it seems like other object references can't re-use the
            # same connection
            self.db.close()
            del SQLITE_CONNECTION_POOL_DB[cache_key(self.db_filename)]
        else:
            # don't close yet because another SQLiteBackingStore is probably still open
            pass

    def _commit(self) -> None:
        self.db.commit()

    def _execute(
        self, sql: str, *args: Optional[List[Any]]
    ) -> Result[Ok[sqlite3.Cursor], Err[str]]:
        with SyftLock(self.lock_config):
            cursor: Optional[sqlite3.Cursor] = None
            err = None
            try:
                cursor = self.cur.execute(sql, *args)
            except Exception as e:
                raise_exception(self.table_name, e)

            # TODO: Which exception is safe to rollback on?
            # we should map out some more clear exceptions that can be returned
            # rather than halting the program like disk I/O error etc
            # self.db.rollback()  # Roll back all changes if an exception occurs.
            # err = Err(str(e))
            self.db.commit()  # Commit if everything went ok

            if err is not None:
                return err

            return Ok(cursor)

    def _set(self, key: UID, value: Any) -> None:
        if self._exists(key):
            self._update(key, value)
        else:
            insert_sql = (
                f"insert into {self.table_name} (uid, repr, value) VALUES (?, ?, ?)"  # nosec
            )
            data = _serialize(value, to_bytes=True)
            res = self._execute(insert_sql, [str(key), _repr_debug_(value), data])
            if res.is_err():
                raise ValueError(res.err())

    def _update(self, key: UID, value: Any) -> None:
        insert_sql = (
            f"update {self.table_name} set uid = ?, repr = ?, value = ? where uid = ?"  # nosec
        )
        data = _serialize(value, to_bytes=True)
        res = self._execute(insert_sql, [str(key), _repr_debug_(value), data, str(key)])
        if res.is_err():
            raise ValueError(res.err())

    def _get(self, key: UID) -> Any:
        select_sql = f"select * from {self.table_name} where uid = ? order by sqltime"  # nosec
        res = self._execute(select_sql, [str(key)])
        if res.is_err():
            raise KeyError(f"Query {select_sql} failed")
        cursor = res.ok()

        row = cursor.fetchone()
        if row is None or len(row) == 0:
            raise KeyError(f"{key} not in {type(self)}")
        data = row[2]
        return _deserialize(data, from_bytes=True)

    def _exists(self, key: UID) -> bool:
        select_sql = f"select uid from {self.table_name} where uid = ?"  # nosec

        res = self._execute(select_sql, [str(key)])
        if res.is_err():
            return False
        cursor = res.ok()

        row = cursor.fetchone()
        if row is None:
            return False

        return bool(row)

    def _get_all(self) -> Any:
        select_sql = f"select * from {self.table_name} order by sqltime"  # nosec
        keys = []
        data = []

        res = self._execute(select_sql)
        if res.is_err():
            return {}
        cursor = res.ok()

        rows = cursor.fetchall()
        if rows is None:
            return {}

        for row in rows:
            keys.append(UID(row[0]))
            data.append(_deserialize(row[2], from_bytes=True))
        return dict(zip(keys, data))

    def _get_all_keys(self) -> Any:
        select_sql = f"select uid from {self.table_name} order by sqltime"  # nosec
        keys = []

        res = self._execute(select_sql)
        if res.is_err():
            return []
        cursor = res.ok()

        rows = cursor.fetchall()
        if rows is None:
            return []

        for row in rows:
            keys.append(UID(row[0]))
        return keys

    def _delete(self, key: UID) -> None:
        select_sql = f"delete from {self.table_name} where uid = ?"  # nosec
        res = self._execute(select_sql, [str(key)])
        if res.is_err():
            raise ValueError(res.err())

    def _delete_all(self) -> None:
        select_sql = f"delete from {self.table_name}"  # nosec
        res = self._execute(select_sql)
        if res.is_err():
            raise ValueError(res.err())

    def _len(self) -> int:
        select_sql = f"select count(uid) from {self.table_name}"  # nosec
        res = self._execute(select_sql)
        if res.is_err():
            raise ValueError(res.err())
        cursor = res.ok()
        cnt = cursor.fetchone()[0]
        return cnt

    def __setitem__(self, key: Any, value: Any) -> None:
        self._set(key, value)

    def __getitem__(self, key: Any) -> Self:
        try:
            return self._get(key)
        except KeyError as e:
            if self._ddtype is not None:
                return self._ddtype()
            raise e

    def __repr__(self) -> str:
        return repr(self._get_all())

    def __len__(self) -> int:
        return self._len()

    def __delitem__(self, key: str):
        self._delete(key)

    def clear(self) -> Self:
        self._delete_all()

    def copy(self) -> Self:
        return deepcopy(self)

    def keys(self) -> Any:
        return self._get_all_keys()

    def values(self) -> Any:
        return self._get_all().values()

    def items(self) -> Any:
        return self._get_all().items()

    def pop(self, key: Any) -> Self:
        value = self._get(key)
        self._delete(key)
        return value

    def __contains__(self, key: Any) -> bool:
        return self._exists(key)

    def __iter__(self) -> Any:
        return iter(self.keys())

    def __del__(self):
        try:
            self._close()
        except BaseException:
            print("Could not close connection")
            pass


@serializable()
class SQLiteStorePartition(KeyValueStorePartition):
    """SQLite StorePartition

    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings, used for indexing and partitioning
        `store_config`: SQLiteStoreConfig
            SQLite specific configuration
    """

    def close(self) -> None:
        self.lock.acquire()
        try:
            # I think we don't want these now, because of the REF_COUNT?
            # self.data._close()
            # self.unique_keys._close()
            # self.searchable_keys._close()
            pass
        except BaseException:
            pass
        self.lock.release()

    def commit(self) -> None:
        self.lock.acquire()
        try:
            self.data._commit()
            self.unique_keys._commit()
            self.searchable_keys._commit()
        except BaseException:
            pass
        self.lock.release()


# the base document store is already a dict but we can change it later
@serializable()
class SQLiteDocumentStore(DocumentStore):
    """SQLite Document Store

    Parameters:
        `store_config`: StoreConfig
            SQLite specific configuration, including connection details and client class type.
    """

    partition_type = SQLiteStorePartition


@serializable()
class SQLiteStoreClientConfig(StoreClientConfig):
    """SQLite connection config

    Parameters:
        `filename` : str
            Database name
        `path` : Path or str
            Database folder
        `check_same_thread`: bool
            If True (default), ProgrammingError will be raised if the database connection is used
            by a thread other than the one that created it. If False, the connection may be accessed
            in multiple threads; write operations may need to be serialized by the user to avoid
            data corruption.
        `timeout`: int
            How many seconds the connection should wait before raising an exception, if the database
            is locked by another connection. If another connection opens a transaction to modify the
            database, it will be locked until that transaction is committed. Default five seconds.
    """

    filename: Optional[str] = None
    path: Union[str, Path] = Field(default_factory=tempfile.gettempdir)
    check_same_thread: bool = True
    timeout: int = 5

    # We need this in addition to Field(default_factory=...)
    # so users can still do SQLiteStoreClientConfig(path=None)
    @validator("path", pre=True)
    def __default_path(cls, path: Optional[Union[str, Path]]) -> Union[str, Path]:
        if path is None:
            return tempfile.gettempdir()
        return path

    @property
    def file_path(self) -> Optional[Path]:
        return Path(self.path) / self.filename if self.filename is not None else None


@serializable()
class SQLiteStoreConfig(StoreConfig):
    __canonical_name__ = "SQLiteStoreConfig"
    """SQLite Store config, used by SQLiteStorePartition

    Parameters:
        `client_config`: SQLiteStoreClientConfig
            SQLite connection configuration
        `store_type`: DocumentStore
            Class interacting with QueueStash. Default: SQLiteDocumentStore
        `backing_store`: KeyValueBackingStore
            The Store core logic. Default: SQLiteBackingStore
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
                * FileLockingConfig: file based locking, ideal for same-device different-processes/threads stores.
                * RedisLockingConfig: Redis-based locking, ideal for multi-device stores.
            Defaults to FileLockingConfig.
    """

    client_config: SQLiteStoreClientConfig
    store_type: Type[DocumentStore] = SQLiteDocumentStore
    backing_store: Type[KeyValueBackingStore] = SQLiteBackingStore
    locking_config: LockingConfig = FileLockingConfig()



# stdlib
from typing import Any
from typing import Optional

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...service.response import SyftError
from ...store.document_store import BaseStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.document_store import UIDPartitionKey
from ...types.uid import UID
from .oblv_keys import OblvKeys


@serializable()
class OblvKeysStash(BaseStash):
    object_type = OblvKeys
    settings: PartitionSettings = PartitionSettings(
        name=OblvKeys.__canonical_name__, object_type=OblvKeys, db_name="app"
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def check_type(self, obj: Any, type_: type) -> Result[Any, str]:
        return (
            Ok(obj)
            if isinstance(obj, type_)
            else Err(f"{type(obj)} does not match required type: {type_}")
        )

    def set(
        self, credentials: SyftVerifyKey, oblv_keys: OblvKeys
    ) -> Result[OblvKeys, Err]:
        if not len(self):
            valid = self.check_type(oblv_keys, self.object_type)
            if valid.is_err():
                return SyftError(message=valid.err())

            return super().set(credentials, oblv_keys)
        else:
            return Err("Domain Node already has an existing public/private key pair")

    def get_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[Optional[OblvKeys], str]:
        qks = QueryKeys(qks=[UIDPartitionKey.with_obj(uid)])
        return Ok(self.query_one(credentials=credentials, qks=qks))

    def delete_by_uid(self, credentials: SyftVerifyKey, uid: UID) -> Result[bool, str]:
        qk = UIDPartitionKey.with_obj(uid)
        return super().delete(qk=qk)

    def update(
        self, credentials: SyftVerifyKey, task: OblvKeys
    ) -> Result[OblvKeys, str]:
        valid = self.check_type(task, self.object_type)
        if valid.is_err():
            return SyftError(message=valid.err())

        return super().update(credentials, task)



# stdlib
from datetime import datetime
from functools import total_ordering
from typing import Optional

# third party
from typing_extensions import Self

# relative
from ..serde.serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .uid import UID


@serializable()
@total_ordering
class DateTime(SyftObject):
    __canonical_name__ = "DateTime"
    __version__ = SYFT_OBJECT_VERSION_1

    id: Optional[UID]
    utc_timestamp: float

    @classmethod
    def now(cls) -> Self:
        return cls(utc_timestamp=datetime.utcnow().timestamp())

    def __str__(self) -> str:
        utc_datetime = datetime.utcfromtimestamp(self.utc_timestamp)
        return utc_datetime.strftime("%Y-%m-%d %H:%M:%S")

    def __hash__(self) -> int:
        return hash(self.utc_timestamp)

    def __eq__(self, other: Self) -> bool:
        return self.utc_timestamp == other.utc_timestamp

    def __lt__(self, other: Self) -> bool:
        return self.utc_timestamp < other.utc_timestamp



# stdlib
from typing import Union

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ..context import AuthedServiceContext
from ..response import SyftError
from ..service import AbstractService
from ..service import service_method
from .object_migration_state import SyftMigrationStateStash
from .object_migration_state import SyftObjectMigrationState


@serializable()
class MigrateStateService(AbstractService):
    store: DocumentStore
    stash: SyftObjectMigrationState

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SyftMigrationStateStash(store=store)

    @service_method(path="migration", name="get_version")
    def get_version(
        self, context: AuthedServiceContext, canonical_name: str
    ) -> Union[int, SyftError]:
        """Search for the metadata for an object."""

        result = self.stash.get_by_name(
            canonical_name=canonical_name, credentials=context.credentials
        )

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        migration_state = result.ok()

        if migration_state is None:
            return SyftError(
                message=f"No migration state exists for canonical name: {canonical_name}"
            )

        return migration_state.current_version

    @service_method(path="migration", name="get_state")
    def get_state(
        self, context: AuthedServiceContext, canonical_name: str
    ) -> Union[bool, SyftError]:
        result = self.stash.get_by_name(
            canonical_name=canonical_name, credentials=context.credentials
        )

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        return result.ok()

    @service_method(path="migration", name="register_migration_state")
    def register_migration_state(
        self,
        context: AuthedServiceContext,
        current_version: int,
        canonical_name: str,
    ) -> Union[SyftObjectMigrationState, SyftError]:
        obj = SyftObjectMigrationState(
            current_version=current_version, canonical_name=canonical_name
        )
        result = self.stash.set(migration_state=obj, credentials=context.credentials)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        return result.ok()



# stdlib
from threading import Lock
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type

# third party
from pymongo.collection import Collection as MongoCollection
from pymongo.database import Database as MongoDatabase
from pymongo.errors import ConnectionFailure
from pymongo.mongo_client import MongoClient as PyMongoClient
from result import Err
from result import Ok
from result import Result

# relative
from ..serde.serializable import serializable
from .document_store import PartitionSettings
from .document_store import StoreClientConfig
from .document_store import StoreConfig
from .mongo_codecs import SYFT_CODEC_OPTIONS


@serializable()
class MongoStoreClientConfig(StoreClientConfig):
    """
    Paramaters:
        `hostname`: optional string
            hostname or IP address or Unix domain socket path of a single mongod or mongos
            instance to connect to, or a mongodb URI, or a list of hostnames (but no more
            than one mongodb URI). If `host` is an IPv6 literal it must be enclosed in '['
            and ']' characters following the RFC2732 URL syntax (e.g. '[::1]' for localhost).
            Multihomed and round robin DNS addresses are **not** supported.
        `port` : optional int
            port number on which to connect
        `directConnection`: bool
            if ``True``, forces this client to connect directly to the specified MongoDB host
            as a standalone. If ``false``, the client connects to the entire replica set of which
            the given MongoDB host(s) is a part. If this is ``True`` and a mongodb+srv:// URI
            or a URI containing multiple seeds is provided, an exception will be raised.
        `maxPoolSize`: int. Default 100
            The maximum allowable number of concurrent connections to each connected server.
            Requests to a server will block if there are `maxPoolSize` outstanding connections
            to the requested server. Defaults to 100. Can be either 0 or None, in which case
            there is no limit on the number of concurrent connections.
        `minPoolSize` : int. Default 0
            The minimum required number of concurrent connections that the pool will maintain
            to each connected server. Default is 0.
        `maxIdleTimeMS`: int
            The maximum number of milliseconds that a connection can remain idle in the pool
            before being removed and replaced. Defaults to `None` (no limit).
        `appname`: string
            The name of the application that created this MongoClient instance. The server will
            log this value upon establishing each connection. It is also recorded in the slow
            query log and profile collections.
        `maxConnecting`: optional int
            The maximum number of connections that each pool can establish concurrently.
            Defaults to `2`.
        `timeoutMS`: (integer or None)
            Controls how long (in milliseconds) the driver will wait when executing an operation
            (including retry attempts) before raising a timeout error. ``0`` or ``None`` means
            no timeout.
        `socketTimeoutMS`: (integer or None)
            Controls how long (in milliseconds) the driver will wait for a response after sending
            an ordinary (non-monitoring) database operation before concluding that a network error
            has occurred. ``0`` or ``None`` means no timeout. Defaults to ``None`` (no timeout).
        `connectTimeoutMS`: (integer or None)
            Controls how long (in milliseconds) the driver will wait during server monitoring when
            connecting a new socket to a server before concluding the server is unavailable.
            ``0`` or ``None`` means no timeout. Defaults to ``20000`` (20 seconds).
        `serverSelectionTimeoutMS`: (integer)
            Controls how long (in milliseconds) the driver will wait to find an available, appropriate
            server to carry out a database operation; while it is waiting, multiple server monitoring
            operations may be carried out, each controlled by `connectTimeoutMS`.
            Defaults to ``120000`` (120 seconds).
        `waitQueueTimeoutMS`: (integer or None)
            How long (in milliseconds) a thread will wait for a socket from the pool if the pool
            has no free sockets. Defaults to ``None`` (no timeout).
        `heartbeatFrequencyMS`: (optional)
            The number of milliseconds between periodic server checks, or None to accept the default
            frequency of 10 seconds.
        # Auth
        username: str
            Database username
        password: str
            Database pass
        authSource: str
            The database to authenticate on.
            Defaults to the database specified in the URI, if provided, or to “admin”.
        tls: bool
            If True, create the connection to the server using transport layer security.
            Defaults to False.
        # Testing and connection reuse
        client: Optional[PyMongoClient]
            If provided, this client is reused. Default = None

    """

    # Connection
    hostname: Optional[str] = "127.0.0.1"
    port: Optional[int] = None
    directConnection: bool = False
    maxPoolSize: int = 200
    minPoolSize: int = 0
    maxIdleTimeMS: Optional[int] = None
    maxConnecting: int = 3
    timeoutMS: int = 0
    socketTimeoutMS: int = 0
    connectTimeoutMS: int = 20000
    serverSelectionTimeoutMS: int = 120000
    waitQueueTimeoutMS: Optional[int] = None
    heartbeatFrequencyMS: int = 10000
    appname: str = "pysyft"
    # Auth
    username: Optional[str] = None
    password: Optional[str] = None
    authSource: str = "admin"
    tls: Optional[bool] = False
    # Testing and connection reuse
    client: Any = None

    # this allows us to have one connection per `Node` object
    # in the MongoClientCache
    node_obj_python_id: Optional[int] = None


class MongoClientCache:
    __client_cache__: Dict[str, Type["MongoClient"]] = {}
    _lock: Lock = Lock()

    @classmethod
    def from_cache(cls, config: MongoStoreClientConfig) -> Optional[PyMongoClient]:
        return cls.__client_cache__.get(hash(str(config)), None)

    @classmethod
    def set_cache(cls, config: MongoStoreClientConfig, client: PyMongoClient) -> None:
        with cls._lock:
            cls.__client_cache__[hash(str(config))] = client


class MongoClient:
    client: PyMongoClient = None

    def __init__(self, config: MongoStoreClientConfig, cache: bool = True) -> None:
        self.config = config
        if config.client is not None:
            self.client = config.client
        elif cache:
            self.client = MongoClientCache.from_cache(config=config)

        if not cache or self.client is None:
            self.connect(config=config)

    def connect(self, config: MongoStoreClientConfig) -> Result[Ok, Err]:
        self.client = PyMongoClient(
            # Connection
            host=config.hostname,
            port=config.port,
            directConnection=config.directConnection,
            maxPoolSize=config.maxPoolSize,
            minPoolSize=config.minPoolSize,
            maxIdleTimeMS=config.maxIdleTimeMS,
            maxConnecting=config.maxConnecting,
            timeoutMS=config.timeoutMS,
            socketTimeoutMS=config.socketTimeoutMS,
            connectTimeoutMS=config.connectTimeoutMS,
            serverSelectionTimeoutMS=config.serverSelectionTimeoutMS,
            waitQueueTimeoutMS=config.waitQueueTimeoutMS,
            heartbeatFrequencyMS=config.heartbeatFrequencyMS,
            appname=config.appname,
            # Auth
            username=config.username,
            password=config.password,
            authSource=config.authSource,
            tls=config.tls,
            uuidRepresentation="standard",
        )
        MongoClientCache.set_cache(config=config, client=self.client)
        try:
            # Check if mongo connection is still up
            self.client.admin.command("ping")
        except ConnectionFailure as e:
            self.client = None
            return Err(str(e))

        return Ok()

    def with_db(self, db_name: str) -> Result[MongoDatabase, Err]:
        try:
            return Ok(self.client[db_name])
        except BaseException as e:
            return Err(str(e))

    def with_collection(
        self,
        collection_settings: PartitionSettings,
        store_config: StoreConfig,
        collection_name: Optional[str] = None,
    ) -> Result[MongoCollection, Err]:
        res = self.with_db(db_name=store_config.db_name)
        if res.is_err():
            return res
        db = res.ok()

        try:
            collection_name = (
                collection_name
                if collection_name is not None
                else collection_settings.name
            )
            collection = db.get_collection(
                name=collection_name, codec_options=SYFT_CODEC_OPTIONS
            )
        except BaseException as e:
            return Err(str(e))

        return Ok(collection)

    def with_collection_permissions(
        self, collection_settings: PartitionSettings, store_config: StoreConfig
    ) -> Result[MongoCollection, Err]:
        """
        For each collection, create a corresponding collection
        that store the permissions to the data in that collection
        """
        res = self.with_db(db_name=store_config.db_name)
        if res.is_err():
            return res
        db = res.ok()

        try:
            collection_permissions_name: str = collection_settings.name + "_permissions"
            collection_permissions = db.get_collection(
                name=collection_permissions_name, codec_options=SYFT_CODEC_OPTIONS
            )
        except BaseException as e:
            return Err(str(e))

        return Ok(collection_permissions)

    def close(self):
        self.client.close()
        MongoClientCache.__client_cache__.pop(hash(str(self.config)), None)



# stdlib

# relative
from ..service.user.user_roles import ServiceRole
from .exception import PySyftException

UserAlreadyExistsException = PySyftException(
    message="User already exists", roles=[ServiceRole.ADMIN]
)



# stdlib
from datetime import datetime
from datetime import timedelta
import mimetypes
from pathlib import Path
from queue import Queue
import sys
import threading
from time import sleep
from typing import Any
from typing import ClassVar
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from azure.storage.blob import BlobSasPermissions
from azure.storage.blob import generate_blob_sas
from botocore.client import ClientError as BotoClientError
from typing_extensions import Self

# relative
from ..node.credentials import SyftVerifyKey
from ..serde import serialize
from ..serde.serializable import serializable
from ..service.action.action_object import ActionObject
from ..service.action.action_object import BASE_PASSTHROUGH_ATTRS
from ..service.action.action_types import action_types
from ..service.response import SyftException
from ..service.service import from_api_or_context
from ..types.grid_url import GridURL
from ..types.transforms import drop
from ..types.transforms import keep
from ..types.transforms import make_set_default
from ..types.transforms import transform
from .datetime import DateTime
from .syft_migration import migrate
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SYFT_OBJECT_VERSION_2
from .syft_object import SyftObject
from .uid import UID

READ_EXPIRATION_TIME = 1800  # seconds
DEFAULT_CHUNK_SIZE = 10000 * 1024


@serializable()
class BlobFileV1(SyftObject):
    __canonical_name__ = "BlobFile"
    __version__ = SYFT_OBJECT_VERSION_1

    file_name: str

    __repr_attrs__ = ["id", "file_name"]


@serializable()
class BlobFile(SyftObject):
    __canonical_name__ = "BlobFile"
    __version__ = SYFT_OBJECT_VERSION_2

    file_name: str
    syft_blob_storage_entry_id: Optional[UID] = None
    file_size: Optional[int] = None

    __repr_attrs__ = ["id", "file_name"]

    def read(self, stream=False, chunk_size=DEFAULT_CHUNK_SIZE, force=False):
        # get blob retrieval object from api + syft_blob_storage_entry_id
        read_method = from_api_or_context(
            "blob_storage.read", self.syft_node_location, self.syft_client_verify_key
        )
        blob_retrieval_object = read_method(self.syft_blob_storage_entry_id)
        return blob_retrieval_object._read_data(stream=stream, chunk_size=chunk_size)

    @classmethod
    def upload_from_path(self, path, client):
        # syft absolute
        import syft as sy

        return sy.ActionObject.from_path(path=path).send(client).syft_action_data

    def _iter_lines(self, chunk_size=DEFAULT_CHUNK_SIZE):
        """Synchronous version of the async iter_lines. This implementation
        is also optimized in terms of splitting chunks, making it faster for
        larger lines"""
        pending = None
        for chunk in self.read(stream=True, chunk_size=chunk_size):
            if b"\n" in chunk:
                if pending is not None:
                    chunk = pending + chunk
                lines = chunk.splitlines()
                if lines and lines[-1] and chunk and lines[-1][-1] == chunk[-1]:
                    pending = lines.pop()
                else:
                    pending = None
                yield from lines
            else:
                if pending is None:
                    pending = chunk
                else:
                    pending = pending + chunk

        if pending is not None:
            yield pending

    def read_queue(self, queue, chunk_size, progress=False, buffer_lines=10000):
        total_read = 0
        for _i, line in enumerate(self._iter_lines(chunk_size=chunk_size)):
            line_size = len(line) + 1  # add byte for \n
            if self.file_size is not None:
                total_read = min(self.file_size, total_read + line_size)
            else:
                # naive way of doing this, max be 1 byte off because the last
                # byte can also be a \n
                total_read += line_size
            if progress:
                queue.put((total_read, line))
            else:
                queue.put(line)
            while queue.qsize() > buffer_lines:
                sleep(0.1)
        # Put anything not a string at the end
        queue.put(0)

    def iter_lines(self, chunk_size=DEFAULT_CHUNK_SIZE, progress=False):
        item_queue: Queue = Queue()
        threading.Thread(
            target=self.read_queue,
            args=(item_queue, chunk_size, progress),
            daemon=True,
        ).start()
        item = item_queue.get()
        while item != 0:
            yield item
            item = item_queue.get()

    def _coll_repr_(self):
        return {"file_name": self.file_name}


@migrate(BlobFile, BlobFileV1)
def downgrade_blobfile_v2_to_v1():
    return [
        drop(["syft_blob_storage_entry_id", "file_size"]),
    ]


@migrate(BlobFileV1, BlobFile)
def upgrade_blobfile_v1_to_v2():
    return [
        make_set_default("syft_blob_storage_entry_id", None),
        make_set_default("file_size", None),
    ]


class BlobFileType(type):
    pass


class BlobFileObjectPointer:
    pass


@serializable()
class BlobFileObject(ActionObject):
    __canonical_name__ = "BlobFileOBject"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_internal_type: ClassVar[Type[Any]] = BlobFile
    syft_pointer_type = BlobFileObjectPointer
    syft_passthrough_attrs = BASE_PASSTHROUGH_ATTRS


@serializable()
class SecureFilePathLocation(SyftObject):
    __canonical_name__ = "SecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    path: str

    def __repr__(self) -> str:
        return f"{self.path}"

    def generate_url(self, *args):
        raise NotImplementedError


@serializable()
class SeaweedSecureFilePathLocationV1(SecureFilePathLocation):
    __canonical_name__ = "SeaweedSecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_1

    upload_id: str


@serializable()
class SeaweedSecureFilePathLocation(SecureFilePathLocation):
    __canonical_name__ = "SeaweedSecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_2

    upload_id: Optional[str] = None

    def generate_url(self, connection, type_, bucket_name):
        try:
            url = connection.client.generate_presigned_url(
                ClientMethod="get_object",
                Params={"Bucket": bucket_name, "Key": self.path},
                ExpiresIn=READ_EXPIRATION_TIME,
            )

            # relative
            from ..store.blob_storage import BlobRetrievalByURL

            return BlobRetrievalByURL(
                url=GridURL.from_url(url), file_name=Path(self.path).name, type_=type_
            )
        except BotoClientError as e:
            raise SyftException(e)


@migrate(SeaweedSecureFilePathLocationV1, SeaweedSecureFilePathLocation)
def upgrade_seaweedsecurefilepathlocation_v1_to_v2():
    return [make_set_default("bucket_name", "")]


@migrate(SeaweedSecureFilePathLocation, SeaweedSecureFilePathLocationV1)
def downgrade_seaweedsecurefilepathlocation_v2_to_v1():
    return [
        drop(["bucket_name"]),
    ]


@serializable()
class AzureSecureFilePathLocation(SecureFilePathLocation):
    __canonical_name__ = "AzureSecureFilePathLocation"
    __version__ = SYFT_OBJECT_VERSION_1

    # upload_id: str
    azure_profile_name: str  # Used by Seaweedfs to refer to a remote config
    bucket_name: str

    def generate_url(self, connection, type_, *args):
        # SAS is almost the same thing as the presigned url
        config = connection.config.remote_profiles[self.azure_profile_name]
        account_name = config.account_name
        container_name = config.container_name
        blob_name = self.path
        sas_blob = generate_blob_sas(
            account_name=account_name,
            container_name=container_name,
            blob_name=blob_name,
            account_key=config.account_key,
            permission=BlobSasPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(hours=48),
        )
        url = f"https://{config.account_name}.blob.core.windows.net/{container_name}/{blob_name}?{sas_blob}"

        # relative
        from ..store.blob_storage import BlobRetrievalByURL

        return BlobRetrievalByURL(url=url, file_name=Path(self.path).name, type_=type_)


@serializable()
class BlobStorageEntryV1(SyftObject):
    __canonical_name__ = "BlobStorageEntry"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    location: Union[SecureFilePathLocation, SeaweedSecureFilePathLocation]
    type_: Optional[Type]
    mimetype: str = "bytes"
    file_size: int
    uploaded_by: SyftVerifyKey
    created_at: DateTime = DateTime.now()

    __attr_searchable__ = ["bucket_name"]


@serializable()
class BlobStorageEntry(SyftObject):
    __canonical_name__ = "BlobStorageEntry"
    __version__ = SYFT_OBJECT_VERSION_2

    id: UID
    location: Union[SecureFilePathLocation, SeaweedSecureFilePathLocation]
    type_: Optional[Type]
    mimetype: str = "bytes"
    file_size: int
    no_lines: Optional[int] = 0
    uploaded_by: SyftVerifyKey
    created_at: DateTime = DateTime.now()
    bucket_name: Optional[str]

    __attr_searchable__ = ["bucket_name"]


@migrate(BlobStorageEntry, BlobStorageEntryV1)
def downgrade_blobstorageentry_v2_to_v1():
    return [
        drop(["no_lines", "bucket_name"]),
    ]


@migrate(BlobStorageEntryV1, BlobStorageEntry)
def upgrade_blobstorageentry_v1_to_v2():
    return [make_set_default("no_lines", 1), make_set_default("bucket_name", None)]


@serializable()
class BlobStorageMetadataV1(SyftObject):
    __canonical_name__ = "BlobStorageMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    type_: Optional[Type[SyftObject]]
    mimetype: str = "bytes"
    file_size: int


@serializable()
class BlobStorageMetadata(SyftObject):
    __canonical_name__ = "BlobStorageMetadata"
    __version__ = SYFT_OBJECT_VERSION_2

    type_: Optional[Type[SyftObject]]
    mimetype: str = "bytes"
    file_size: int
    no_lines: Optional[int] = 0


@migrate(BlobStorageMetadata, BlobStorageMetadataV1)
def downgrade_blobmeta_v2_to_v1():
    return [
        drop(["no_lines"]),
    ]


@migrate(BlobStorageMetadataV1, BlobStorageMetadata)
def upgrade_blobmeta_v1_to_v2():
    return [make_set_default("no_lines", 1)]


@serializable()
class CreateBlobStorageEntry(SyftObject):
    __canonical_name__ = "CreateBlobStorageEntry"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    type_: Optional[Type]
    mimetype: str = "bytes"
    file_size: int
    extensions: List[str] = []

    @classmethod
    def from_obj(cls, obj: SyftObject) -> Self:
        file_size = sys.getsizeof(serialize._serialize(obj=obj, to_bytes=True))
        return cls(file_size=file_size, type_=type(obj))

    @classmethod
    def from_path(cls, fp: Union[str, Path], mimetype: Optional[str] = None) -> Self:
        path = Path(fp)
        if not path.exists():
            raise SyftException(f"{fp} does not exist.")
        if not path.is_file():
            raise SyftException(f"{fp} is not a file.")

        if fp.suffix.lower() == ".jsonl":
            mimetype = "application/json-lines"
        if mimetype is None:
            mime_types = mimetypes.guess_type(fp)
            if len(mime_types) > 0 and mime_types[0] is not None:
                mimetype = mime_types[0]
            else:
                raise SyftException(
                    "mimetype could not be identified.\n"
                    "Please specify mimetype manually `from_path(..., mimetype = ...)`."
                )

        return cls(
            mimetype=mimetype,
            file_size=path.stat().st_size,
            extensions=path.suffixes,
            type_=BlobFileType,
        )

    @property
    def file_name(self) -> str:
        return str(self.id) + "".join(self.extensions)


@transform(BlobStorageEntry, BlobStorageMetadata)
def storage_entry_to_metadata():
    return [keep(["id", "type_", "mimetype", "file_size"])]


action_types[BlobFile] = BlobFileObject
